// dosepre.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 15:06:55 tac>
//
// クロス集計表を入力し、プロビット分析による用量反応解析を行う。
// クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W dosepre dosepre.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "dosepre  Version 7.41e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:dosepre [<OPTION> ...] [<DATA FILE> ...]",
  "  The dose-response analysis on a cross table with vertical as dose items",
  "  and horizontal as response items.",
  "<OPTION>",
  "  -du={l|r}:select the unit of dose(default:-du=l)",
  "  -eps[=<ERROR>]:specify the convergence error of the maximum likelihood",
  "    method(0<ERROR)",
  "  -f=<FILE>:reads options from the specified file",
  "  -g[={-|+}]:specify whether to use the Fieller's formula(-g:-g=+)",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -log[={c|w}]:select the type of logit analysis(-log:-log=l)",
  "    c:repetitive completion method(default)  w:working logit method",
  "  -ngp={drc|drl|drlp|pro}[,...]:select the data for Ngraph",
  "    drc:dose-response curve of probit analysis",
  "    drl:dose-response line of trend analysis",
  "    drlp:dose-response curve of logit analysis",
  "    pro:response rate  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output",
  "    -ngpname:-ngpname=dosepre_ngp<2-digit serial number>.dat",
  "  -res=[-][{def|all|ca|cc[<C-COEF>]|log|pa|pr[<RATE>]|r",
  "      |tab[{fr|pc|pr|pt}]}][,...]:select output results",
  "    -:wiht -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    ca:trend analysis",
  "    cc[<C-COEF>]:confidence interval(0=<C-COEF<1 cc:cc0.95)",
  "    log:logit analysis  pa:probit analysis",
  "    pr[<RATE>]:inverse estimation dose for response rate(0=<RATE<1)",
  "    r:contribution rate  tab[{fr|pc|pr|pt}]:cross table",
  "      fr:only frequancy  pc:also output % for column total",
  "      pr:also output % for row total  pt:also output % of all total",
  "  -rmax=<NUMBER>:specify the maximum number of repetitions(0=<NUMBER)"
};

const string Title = "=== Dose-Response Analysis of Nominal Scale ==="; // 表題
const string NGPNAMEdef = "dosepre_ngp"; // 省略時Ngraphファイル名
constexpr double EPSdef = 0.00001; // 省略時収束誤差
constexpr double PRdef = 0.5;      // 省略時用量推定反応率

enum dutype                     // 用量単位の種類
{
  dupnon,                       // 単位なし
  dulog,                        // 対数
  dureal,                       // 実数
  duetc                         // その他
};

enum logtype                    // ロジット分析の種類
{
  lognon,                       // 実行せず
  logcomp,                      // 繰り返し補完法
  logwork,                      // 実用ロジット
  logetc                        // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  dutype vdu;                   // 用量の単位
  double veps;                  // 最尤法の収束誤差
  bool vg;                      // Fiellerの式使用フラッグ
  logtype vlog;                 // ロジット分析の種類
  string vngpname;              // Ngraphファイル基本名
  string vngpnamen;             // Ngraphファイル名(番号付き)
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  double vpr;                   // 用量を推定する反応率
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  int vrmax;                    // 最大反復回数
  tabtype vtab;                 // クロス集計表の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  dutype du() const             // 用量の単位を返す
  { return vdu; }
  double eps() const            // 最尤法の収束誤差を返す
  { return veps; }
  bool g() const                // Fiellerの式使用フラッグをす
  { return vg; }
  logtype log() const           // ロジット分析の種類を返す
  { return vlog; }
  const string& ngpname(int fno); // Ngraphデータファイル名を返す
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  double pr() const             // 用量を推定する反応率を返す
  { return vpr; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  int rmax() const              // 最大反復回数を返す
  { return vrmax; }
  tabtype tab() const           // クロス集計表の種類を返す
  { return vtab; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check();                  // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdu(dulog), veps(EPSdef), vg(true), vlog(logcomp), vngpname(), vngpnamen(), vngplst(), vngpmap(), vpr(PRdef), vreslst(), vresmap(), vrmax(0), vtab(tabfprr)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

const string& option::ngpname(int fno) // Ngraphデータファイル名を返す
{
  if (vngpname == "") return vngpname; // -ngpname指定なし
  ostringstream ostr;
  if (fno < 10) ostr << vngpname << "0" << fno << ".dat";
  else ostr << vngpname << fno << ".dat";
  vngpnamen = (string)(ostr.str());
  return vngpnamen;             // 番号付きファイル名を返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "du")           // 用量の単位指定
  {
    if (opt[1].empty() || opt[1] == "l") vdu = dulog;
    else if (opt[1] == "r") vdu = dureal;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "eps")     // 最尤法の収束誤差の指定
  {
    veps = opt[1].empty() ? EPSdef : strtodf(opt[1]);
    if (veps < DBL_EPSILON) err = badopt(vpname, optstr, USAGE, DBL_EPSILON, DBL_MAX);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "g")       // Fiellerの式使用指定
  {
    if (opt[1].empty() || opt[1] == "+") vg = true;
    else if (opt[1] == "-") vg = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "log")     // ロジット分析の種類指定
  {
    if (opt[1].empty() || opt[1] == "c") vlog = logcomp;
    else if (opt[1] == "w") vlog = logwork;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
  }
  else if (opt[0] == "pr")       // 用量を推定する反応率の指定
  {
    vpr = opt[1].empty() ? PRdef : strtodf(opt[1]);
    if (vpr <= 0.0 || 1.0 <= vpr) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "rmax")    // 最大反復回数指定
  {
    char* ptr = nullptr;
    vrmax = strtol(opt[1].c_str(), &ptr, 0);
    if (vrmax < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check()             // オプションのチェックとセット
{
  if (set_ngp() < 0) return -2; // -resオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["drc"] = vngpmap["drl"] = vngpmap["drlp"] = vngpmap["pro"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["drc"] = vngpmap["drl"] = vngpmap["drlp"] = vngpmap["pro"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["pa"] = vresmap["pr"] = true;
  vresmap["ca"] = vresmap["log"] = vresmap["r"] = vresmap["tab"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["pa"] = vresmap["pr"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["pa"] = vresmap["pr"] = true;
      }
      else if (res == "all")
      {
        vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["pa"] = vresmap["pr"] = vresmap["ca"] = vresmap["log"] = vresmap["r"] = vresmap["tab"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res[0] == 'p' && res[1] == 'r' && res.size() > 2)
      {
        vpr = strtodf(res.substr(2)); // 反応率
        if (vpr < 0.0 || 1.0 <= vpr) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "tabfr")
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (res == "tabpc")
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (res == "tabpr")
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (res == "tabpt")
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class npro                      // 用量別プロビットクラス
{
private:
  double vd;                    // 用量
  double vx;                    // 計算用用量
  int vtn;                      // 総例数
  int vrn;                      // 反応例数
  double vpc;                   // 補正反応率
  double vyo;                   // 観測プロビット
  double vy;                    // 期待プロビット
public:
  npro() : vd(0.0), vx(0.0), vtn(0), vrn(0), vpc(0.0), vyo(0.0), vy(0.0)
  // デフォールトコンストラクタ
  {}
  npro(double d0, double x0, int tn0, int rn0) : vd(d0), vx(x0), vtn(tn0), vrn(rn0), vpc(0.0), vyo(0.0), vy(0.0)
  // コンストラクタ
  {}
  npro(const npro& npro0) : vd(npro0.vd), vx(npro0.vx), vtn(npro0.vtn), vrn(npro0.vrn), vpc(npro0.vpc), vyo(npro0.vyo), vy(npro0.vy)
  // コピーコンストラクタ
  {}
  virtual ~npro()               // デストラクタ
  {}
  const npro& operator=(const npro& npro0); // コピー代入演算子
  double d() const                          // 用量を返す
  { return vd; }
  double d(double d0)           // 用量を設定する
  { return (vd = d0); }
  double x() const              // 計算用用量を返す
  { return vx; }
  double x(double x0)           // 計算用用量を設定する
  { return (vx = x0); }
  int tn() const                // 総例数を返す
  { return vtn; }
  int tn(int tn0)               // 総例数を設定する
  { return (vtn = tn0); }
  int rn() const                // 反応例数を返す
  { return vrn; }
  int rn(int rn0)               // 反応例数を設定する
  { return (vrn = rn0); }
  double p() const              // 反応率を返す
  { return (vtn > 0) ? (double)vrn / (double)vtn : 0.0; }
  double pc() const             // 補正反応率を返す
  { return vpc; }
  double yo() const             // 観測プロビットを返す
  { return vyo; }
  double yo(double yo0)         // 観測プロビットを設定する
  { return (vyo = yo0); }
  double y() const              // 期待プロビットを返す
  { return vy; }
  double y(double y0)           // 期待プロビットを設定する
  { return (vy = y0); }
  void sigma(int tn0, int rn0)  // データを累積する
  { vtn += tn0;  vrn += rn0; }
  double calpc(double p0);      // 補正反応率を計算する
};

const npro& npro::operator=(const npro& npro0) // コピー代入演算子
{
  if (&npro0 == this) return *this; // 自分自身への代入を防ぐ
  vd = npro0.vd;  vx = npro0.vx;  vtn = npro0.vtn;  vrn = npro0.vrn;
  vpc = npro0.vpc;  vyo= npro0.vyo;  vy = npro0.vy;
  return *this;
}

double npro::calpc(double p0)   // 補正反応率を計算する
{
  vpc = (vtn > 0) ? (double)vrn / (double)vtn : 0.0;
  if (p0 > 0.0) vpc = (vpc - p0) / (1.0 - p0);
  if (vpc < 0.0) vpc = 0.0;
  return vpc;                   // 補正反応率を返す
}

class varclass                  // 変数クラス
{
private:
  int vnd;                      // 用量数
  vector<npro> nprovec;         // 用量別プロビットベクトル
  statcor<double> dstatc;       // 用量プロビット用相関統計量
  double vsnw;                  // 重み合計
  double vmxw;                  // xの重み付き平均値
  double vmyw;                  // yの重み付き平均値
  double vsxxw;                 // xの重み付き平方和
  double vsyyw;                 // yの重み付き平方和
  double vsxyw;                 // xとyの重み付き積和
  double va;                    // 用量プロビット直線の定数
  double vb;                    // 用量プロビット直線の回帰係数
  int vnr;                      // 繰り返し計算数
  void cal_pline(const option& opt); // プロビット直線の近似計算
  int cal_logit(double& a, double& b, const option& opt, bool comp);
  // 用量-ロジット分析
public:
  varclass(int nd0) : vnd(nd0), nprovec(nd0), dstatc(), vsnw(0.0), vmxw(0.0), vsxxw(0.0), vsyyw(0.0), vsxyw(0.0), va(0.0), vb(0.0), vnr(0) // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int cal(const option& opt, const ftable& ftab);
  // 用量別反応率とプロビットを計算する
  int caldpline(const option& opt);
  // Newton法による用量-プロビット直線の反復計算
  int calcaline(const option& opt);
  // Cochran-Armitageの用量-反応率直線の計算
  int callpline(const option& opt); // 用量-ロジット直線の反復計算
  int nd() const                    // 用量数を返す
  { return vnd; }
  double d(size_t idx) const    // 用量を返す
  { return (idx < nprovec.size()) ? nprovec[idx].d() : 0.0; }
  double x(size_t idx) const    // 計算用用量を返す
  { return (idx < nprovec.size()) ? nprovec[idx].x() : 0.0; }
  int tn(size_t idx) const      // 総例数を返す
  { return (idx < nprovec.size()) ? nprovec[idx].tn() : 0.0; }
  int rn(size_t idx) const      // 反応例数を返す
  { return (idx < nprovec.size()) ? nprovec[idx].rn() : 0.0; }
  double p(size_t idx) const    // 反応率を返す
  { return (idx < nprovec.size()) ? nprovec[idx].p() : 0.0; }
  double pc(size_t idx) const   // 補正反応率を返す
  { return (idx < nprovec.size()) ? nprovec[idx].pc() : 0.0; }
  double yo(size_t idx) const   // 観測プロビットを返す
  { return (idx < nprovec.size()) ? nprovec[idx].yo() : 0.0; }
  double y(size_t idx) const    // 期待プロビットを返す
  { return (idx < nprovec.size()) ? nprovec[idx].y() : 0.0; }
  double snw() const            // 重み合計を返す
  { return vsnw; }
  double mxw() const            // xの重み付き平均値を返す
  { return vmxw; }
  double myw() const            // yの重み付き平均値を返す
  { return vmyw; }
  double sxxw() const           // xの重み付き平方和を返す
  { return vsxxw; }
  double syyw() const           // yの重み付き平方和を返す
  { return vsyyw; }
  double sxyw() const           // xとyの重み付き積和を返す
  { return vsxyw; }
  double a() const              // 用量プロビット直線の定数を返す
  { return va; }
  double b() const              // 用量プロビット直線の回帰係数を返す
  { return vb; }
  int nr() const                // 繰り返し計算数を返す
  { return vnr; }
};

void varclass::cal_pline(const option& opt) // プロビット直線の近似計算
{
  constexpr double Sqrpi = 1.0 / sqrt(2.0 * M_PI);
  vsnw = vmxw = vmyw = vsxxw = vsyyw = vsxyw = 0.0;
  double p0 = 0.0;
  int istart = 0;
  if (opt.du() == dulog && nprovec[0].d() == 0.0)
  {
    p0 = nprovec[0].p();
    istart = 1;
  }
  double pc = p0 / (1.0 - p0);
  for (int i = istart; i < vnd; ++i)
  {
    double x = nprovec[i].x(), y = nprovec[i].y();
    if (x != BLANKf && y != BLANKf)
    {
      double ned = (y - 5.0) * (y - 5.0);
      if (ned > 50.0) ned = 50.0;
      double z = Sqrpi * exp(-ned * 0.5);
      if ((ned = y - 5.0) < -10.0) ned = -10.0;
      double p = nedp(ned);
      y += (nprovec[i].pc() - p) / z;
      double nw = nprovec[i].tn() * z * z / (p + pc) / (1.0 - p);
      vsnw += nw;  vmxw += nw * x;  vsxxw += nw * x * x;
      vmyw += nw * y;  vsyyw += nw * y * y;  vsxyw += nw * x * y;
    }
  }
  double eps = DBL_EPSILON * fmax(5.0, vsyyw);
  if (vsnw < eps)
  {
    vsnw = vmxw = vmyw = vsxxw = vsyyw = vsxyw = 0.0;
    return;
  }
  vsxxw = vsxxw - vmxw * vmxw / vsnw;
  if (vsxxw < eps) vsxxw = 0.0;
  vsyyw = vsyyw - vmyw * vmyw / vsnw;
  if (vsyyw < eps) vsyyw = 0.0;
  vsxyw = vsxyw - vmxw * vmyw / vsnw;
  if (fabs(vsxyw) < eps) vsxyw = 0.0;
  vmxw = vmxw / vsnw;  vmyw = vmyw / vsnw;
}

int varclass::cal_logit(double& a, double& b, const option& opt, bool comp)
// 用量-ロジット分析
{
  vsnw = vmxw = vmyw = vsxxw = vsyyw = vsxyw = 0.0;
  int nest = 0;                 // 推測値で補完した用量数
  double y = 0.0, nw = 0.0;
  int istart = 0;
  if (opt.du() == dulog && nprovec[0].d() == 0.0) istart = 1; // 用量0のデータがある時
  for (int i = istart; i < vnd; ++i)
  {
    double x = nprovec[i].x(), pc = nprovec[i].pc();
    int n = nprovec[i].tn();
    if (0.0 < pc && pc < 1.0)
    {
      y = log(pc / (1.0 - pc));
      nw = (double)n * pc * (1.0 - pc);
    }
    else
    {
      if (comp)                 // 推測値で補完する
      {
        double p = 1.0 / (1.0 + exp(-a - b * x));
        y = log(p / (1.0 - p));  nw = (double)n * p * (1.0 - p);
      }
      else                      // 実用ロジットを用いる
      {
        double cf = 0.5 / (double)n;
        y = log((pc + cf) / (1.0 - pc + cf));
        nw = (double)n * (pc + cf) * (1.0 - pc + cf) / (1.0 + 2.0 * cf);
      }
      ++nest;
    }
    vsnw += nw;  vmxw += nw * x;  vsxxw += nw * x * x;
    vmyw += nw * y;  vsyyw += nw * y * y;  vsxyw += nw * x * y;
  }
  const double Seps = DBL_EPSILON * fmax(5.0, vsyyw);
  if (vsnw < Seps)
  {
    vsnw = vmxw = vmyw = vsxxw = vsyyw = vsxyw = 0.0;
    return -1;                  // ロジット分析不可能
  }
  vsxxw = vsxxw - vmxw * vmxw / vsnw;
  if (vsxxw < Seps) vsxxw = 0.0;
  vsyyw = vsyyw - vmyw * vmyw / vsnw;
  if (vsyyw < Seps) vsyyw = 0.0;
  vsxyw = vsxyw - vmxw * vmyw / vsnw;
  if (fabs(vsxyw) < Seps) vsxyw = 0.0;
  vmxw = vmxw / vsnw;  vmyw = vmyw / vsnw;
  b = vsxyw / vsxxw;
  a = vmyw - b * vmxw;
  return nest;                  // 推測値で補完した用量数を返す
}

int varclass::cal(const option& opt, const ftable& ftab)
// 用量別反応率とプロビットを計算する
{
  double p0 = 0.0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    size_t j = i - 1;
    double dose = strtodf(ftab.rcode(i));
    nprovec[j].d(dose);
    double x = dose;
    if (opt.du() == dulog) x = (dose > 0.0) ? log10(dose) : BLANKf;
    nprovec[j].x(x);
    nprovec[j].tn(ftab.frtotal(i));  nprovec[j].rn(ftab.f(i, 2));
    if (i == 1 && opt.du() == dulog && dose == 0.0) // 用量0のデータがある時
    {
      p0 = nprovec[j].p();      // 反応率補正用
      nprovec[j].calpc(p0);     // 補正反応率計算(0になる)
      nprovec[j].yo(BLANKf);    // 用量0の観測プロビットは欠損にする
      nprovec[j].y(BLANKf);     // 用量0の期待プロビットも欠損にする
    }
    else
    {
      double yo = probit(nprovec[j].calpc(p0)); // 観測プロビット計算
      nprovec[j].yo(yo);
      if (x != BLANKf && yo != BLANKf) dstatc.sigma(x, yo);
    }
  }
  if (dstatc.n() < 2) return dstatc.n(); // 期待プロビット計算できず
  double b = (dstatc.sxx() > 0.0) ? dstatc.sxy() / dstatc.sxx() : 0.0;
  double a = dstatc.my() - b * dstatc.mx();
  int istart = 0;
  if (opt.du() == dulog && nprovec[0].d() == 0.0) istart = 1;
  for (int i = istart; i < vnd; ++i) // 期待プロビットをセットする
  { nprovec[i].y(a + b * nprovec[i].x()); }
  return dstatc.n();            // プロビットが計算できた用量数を返す
}

int varclass::caldpline(const option& opt)
// Newton法による用量-プロビット直線の反復計算
{
  vnr = 0;
  double emax = 0.0;
  do {
    cal_pline(opt);
    vb = (vsxxw > 0.0) ? vsxyw / vsxxw : 0.0;
    va = vmyw - vb * vmxw;
    emax = 0.0;
    for (int i = 0; i < vnd; ++i)
    {
      double x = nprovec[i].x();
      if (x != BLANKf)
      {
        double y = va + vb * x;
        double e = fabs(y - nprovec[i].y());
        if (e > emax) emax = e;
        nprovec[i].y(y);
      }
    }
    ++vnr;
  } while (emax >= opt.eps() && (opt.rmax() < 1 || vnr < opt.rmax()));
  return vnr;                   // 反復計算回数を返す
}

int varclass::calcaline(const option& opt)
// Cochran-Armitageの用量-反応率直線の計算
{
  vsnw = vmxw = vmyw = vsxxw = vsyyw = vsxyw = 0.0;
  int istart = 0;
  if (opt.du() == dulog && nprovec[0].d() == 0.0) istart = 1; // 用量0のデータがある時
  for (int i = istart; i < vnd; ++i)
  {
    double x = nprovec[i].x(), pc = nprovec[i].pc();
    double nw = (double)(nprovec[i].tn());
    vsnw += nw;  vmxw += nw * x;  vmyw += nw * pc;
    vsxxw += nw * x * x;  vsyyw += nw * pc * pc;
    vsxyw += nw * x * pc;
  }
  double eps = DBL_EPSILON * fmax(5.0, vsyyw);
  if (vsnw < eps)               // 計算できず
  {
    vsnw = 0.0;
    return 0;
  }
  vmxw /= vsnw;  vmyw /= vsnw;
  double w = 1.0 / vmyw / (1.0 - vmyw);
  vsnw *= w;
  vsxxw = w * vsxxw - vsnw * vmxw * vmxw;
  if (vsxxw < eps) vsxxw = 0.0;
  vsyyw = w * vsyyw - vsnw * vmyw * vmyw;
  if (vsyyw < eps) vsyyw = 0.0;
  vsxyw = w * vsxyw - vsnw * vmxw * vmyw;
  if (vsxyw < eps) vsxyw = 0.0;
  vb = (vsxxw > 0.0) ? vsxyw / vsxxw : 0.0;
  va = vmyw - vb * vmxw;
  return 1;                     // 計算実行
}

int varclass::callpline(const option& opt) // 用量-ロジット直線の反復計算
{
  vnr = 0;
  int nest = cal_logit(va, vb, opt, false);
  const double Eps = vsyyw * opt.eps();
  if (nest > 0 && opt.log() == logcomp) // 繰り返し補完法
  {
    double syywpre = vsyyw;
    nest = 0;
    bool comp = true;
    while (comp)
    {
      nest = cal_logit(va, vb, opt, true);
      ++vnr;
      if (nest < 0 || fabs(vsyyw - syywpre) < Eps || (opt.rmax() > 0 && vnr >= opt.rmax())) comp = false;
      else syywpre = vsyyw;
    }
    double at = va, bt = vb;    // 収束後のaとbで平方和を求める
    nest = cal_logit(at, bt, opt, true);
  }
  if (nest < 0) return -1;      // 計算できず
  return vnr;                   // 繰り返し数を返す
}

class doseprclass               // doseprメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_dosepr(ostream& fout, option& opt, const ftable& ftab, varclass& var) const;
  // 各種の用量反応解析の結果を出力する
  void put_nstat(ostream& fout, const option& opt, const varclass& var) const;
  // 用量別反応率・プロビット等を出力する
  void put_probit(ostream& fout, option& opt, const varclass& var) const;
  // プロビット分析の結果を出力する
  void put_estdp(ostream& fout, option& opt, const varclass& var) const;
  // プロビット分析による用量の逆推定値を出力する
  void put_doseprtab(ostream& fout, const string& head, double syyw, double sb, int dft) const;
  // 用量反応解析の分散分析表を出力する
  void put_catest(ostream& fout, option& opt, const varclass& var) const;
  // Cochran-Armitageの傾向分析の結果を出力する
  void put_estca(ostream& fout, option& opt, const varclass& var) const;
  // 傾向分析による用量の逆推定値を出力する
  void put_logit(ostream& fout, option& opt, const varclass& var) const;
  // ロジット分析の結果を出力する
  void put_estlog(ostream& fout, option& opt, const varclass& var) const;
  // ロジット分析による用量の逆推定値を出力する
  void plot_ngp(ofstream& fout, option& opt, idd& id, varclass& var) const;
  // Ngraph用データ出力
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngppro(ofstream& fout, const varclass& var) const;
  // Ngraph用に反応率を出力する
  void plot_ngpdrc(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用にプロビット分析の用量-反応率曲線を出力する
  void plot_ngpdrl(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に傾向分析の用量-反応率直線を出力する
  void plot_ngpdrlp(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用にロジット分析の用量-反応率曲線を出力する
public:
  doseprclass() : fname()       // コンストラクタ
  {}
  ~doseprclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // クロス集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst, int fno) const;
  // 結果を出力する
};

void doseprclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "DOSE ITEM(ROW)    :" << id.item(i) << "\n";
      break;
    default : fout << "RESPONSE ITEM(COL):" << id.item(i) << "\n";
    }
  }
}

void doseprclass::put_dosepr(ostream& fout, option& opt, const ftable& ftab, varclass& var) const
// 各種の用量反応解析の結果を出力する
{
  if (opt.res("tab")) putftable(fout, ftab, opt.tab());
  var.caldpline(opt);
  put_nstat(fout, opt, var);
  if (opt.res("pa")) put_probit(fout, opt, var); // プロビット分析出力
  if (opt.res("ca"))            // Cochran-Armitageの傾向分析出力
  {
    var.calcaline(opt);
    put_catest(fout, opt, var);
  }
  if (opt.res("log"))           // ロジット分析出力
  {
    var.callpline(opt);
    put_logit(fout, opt, var);
  }
}

void doseprclass::put_nstat(ostream& fout, const option& opt, const varclass& var) const
// 用量別反応率・プロビット等を出力する
{
  const int LENline = (opt.du() == dureal) ? 68 : 78;
  fout << "\n    Dose";
  if (opt.du() == dulog) fout << " log(Dose)";
  fout << "   Res Rate(Nres/Ntotal)Cor-Res Rate  Obs-Probit  Exp-Probit\n";
  putline(fout, LENline);
  int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
  int istart = 0;
  if (opt.du() == dulog && var.d(0) == 0.0) // 用量0のデータがある時
  {
    fout << setw(8) << setprecision(4) << var.d(0) << setprecision(prec0);
    fout << setw(10) << "-";
    fout << setw(11) << var.p(0);
    fout << "(" << setw(5) << var.rn(0);
    fout << "/";
    fout << left << setw(5) << var.tn(0) << right;
    fout << ")" << setw(12) << 0;
    fout << "\n";
    istart = 1;
  }
  string tstr = "-";
  int npad = tstr.size() - strclen(tstr, CHARwide);
  string tup("-");
  int npadu = tup.size() - strclen(tup, CHARwide);
  for (int i = istart; i < var.nd(); ++i)
  {
    fout << setw(8) << setprecision(4) << var.d(i) << setprecision(prec0);
    if (opt.du() == dulog)
    {
      fout << setw(10) << setprecision(4) << var.x(i) << setprecision(prec0);
    }
    fout << setw(11) << var.p(i);
    fout << "(" << setw(5) << var.rn(i);
    fout << "/" << left << setw(5) << var.tn(i) << right;
    fout << ")" << setw(12) << var.pc(i);
    if (var.pc(i) <= 0.0) fout << setw(12 + npad) << tstr;
    else if (var.pc(i) >= 1.0) fout << setw(12 + npadu) << tup;
    else fout << setw(12) << var.yo(i);
    fout << setw(12) << var.y(i) << "\n";
  }
  putline(fout, LENline);
}

void doseprclass::put_probit(ostream& fout, option& opt, const varclass& var) const
// プロビット分析の結果を出力する
{
  fout << "\n";
  if (opt.du() == dulog) fout << "log(Dose)";
  else fout << "Dose";
  fout << "-Probit Line:y=" << var.a() << showpos << var.b() << noshowpos;
  fout << "x (repetitions=" << var.nr() << ")\n";
  double sb = var.b() * var.sxyw();
  if (opt.res("r"))
  {
    double r2 = sb / var.syyw();
    fout << "Contribution Rate:r-squared=" << r2 << "\n";
  }
  if (opt.res("pr")) put_estdp(fout, opt, var);
  int dft = var.nd() - 1;
  if (opt.du() == dulog && var.d(0) == 0.0) --dft;
  put_doseprtab(fout, "Probit Analysis Table", var.syyw(), sb, dft);
}

void doseprclass::put_estdp(ostream& fout, option& opt, const varclass& var) const
// プロビット分析による用量の逆推定値を出力する
{
  if (sgn(var.b()) == 0) return; // 逆推定できず
  double pr = opt.pr() * 100.0, y = probit(opt.pr());
  double m = (y - var.a()) / var.b();
  fout << "Dose at a Response Rate=" << pr << "%(y=" << y << "):D" << pr << "=";
  if (opt.du() == dureal) fout << "x=" << m << "\n";
  else fout << pow(10.0, m) << "(x=log(D" << pr << ")=" << m << ")\n";
  if (opt.res("cc") == false) return;
  double vr = 1.0, t = ptoz(1.0 - opt.cc());
  double g = t * t * vr / var.b() / var.b() / var.sxxw();
  if (opt.g() == false || g >= 1.0) g = 0.0; // g>1の時は計算できないため
  double gr = 1.0 - g;
  double v = vr * (gr / var.snw() + (m - var.mxw()) * (m - var.mxw()) / var.sxxw());
  double ci = t / var.b() * sqrt(v);
  double ml = (m - g * var.mxw() - ci) / gr, mu = (m - g * var.mxw() + ci) / gr;
  fout << opt.cc() * 100.0 << "%CI of D" << pr << ":";
  if (opt.du() == dureal)
  { fout << "lower=" << ml << " upper=" << mu << "\n"; }
  else
  {
    fout << "lower=" << pow(10.0, ml) << "(x=" << ml;
    fout << ")  upper=" << pow(10.0, mu) << "(x=" << mu << ")\n";
  }
}

void doseprclass::put_doseprtab(ostream& fout, const string& head, double syyw, double sb, int dft) const
// 用量反応解析の分散分析表を出力する
{
  constexpr int LENline = 68;
  fout << "\n" << setw((LENline - strclen(head, CHARwide)) / 2 + head.size()) << head;
  fout << "\nFACTOR      Sum of Squares        DF   Chi-squared       p-value\n";
  putline(fout, LENline);
  int dfb = 1;
  int dfl = dft - dfb;
  double sl = syyw - sb;
  double eps = DBL_EPSILON * fmax(5.0, syyw);
  if (dfl < 1 || sl < eps) sl = 0.0;
  ldbl pr = ctop(sb, dfb);
  fout << "LINEARITY" << setw(17) << sb << setw(10) << dfb;
  fout << setw(14) << sb << setw(14) << pr << pind(pr) << "\n";
  pr = ctop(sl, dfl);
  fout << "HETEROGENEITY" << setw(13) << sl << setw(10) << dfl;
  fout << setw(14) << sl << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  fout << "TOTAL" << setw(21) << syyw << setw(10) << dft << endl;
}

void doseprclass::put_catest(ostream& fout, option& opt, const varclass& var) const
// Cochran-Armitageの傾向分析の結果を出力する
{
  int dft = var.nd() - 1;
  if (opt.du() == dulog && var.d(0) == 0.0) --dft; // 用量0のデータがある時
  double sb = var.sxyw() * var.sxyw() / var.sxxw();
  fout<< "\n";
  if (opt.du() == dulog) fout << "log(Dose)";
  else fout << "Dose";
  fout << "-Response Line:p=" << var.a() << showpos << var.b() << noshowpos << "x\n";
  if (opt.res("r"))
  {
    double r2 = (var.syyw() > 0.0) ? sb / var.syyw() : 0.0;
    fout << "Contribution Rate:r-squared=" << r2 << "\n";
  }
  if (opt.res("pr")) put_estca(fout, opt, var);
  put_doseprtab(fout, "Cochran-Armitage's Trend Analysys Table", var.syyw(), sb, dft);
}

void doseprclass::put_estca(ostream& fout, option& opt, const varclass& var) const
// 傾向分析による用量の逆推定値を出力する
{
  if (sgn(var.b()) == 0) return; // 逆推定できず
  double pr = opt.pr() * 100.0, p = opt.pr();
  double m = (p - var.a()) / var.b();
  fout << "Dose at a Response Rate=" << pr << "%(p=" << p << "):D" << pr << "=";
  if (opt.du() == dureal) fout << "x=" << m << "\n";
  else fout << pow(10.0, m) << "(x=log(D" << pr << ")=" << m << ")\n";
  if (opt.res("cc") == false) return;
  double vr = 1.0, t = ptoz(1.0 - opt.cc());
  double g = t * t * vr / var.b() / var.b() / var.sxxw();
  if (g >= 1.0) g = 0.0;        // g>1の時は計算できないため
  double gr = 1.0 - g;
  double v = vr * (gr / var.snw() + (m - var.mxw()) * (m - var.mxw()) / var.sxxw());
  double ci = t / var.b() * sqrt(v);
  double ml = (m - g * var.mxw() - ci) / gr, mu = (m - g * var.mxw() + ci) / gr;
  fout << opt.cc() * 100.0 << "% CI of D" << pr << ":";
  if (opt.du() == dureal)
  { fout << "lower=" << ml << "  upper=" << mu << "\n"; }
  else
  {
    fout << "lower=" << pow(10.0, ml) << "(x=" << ml;
    fout << ")  upper=" << pow(10.0, mu) << "(x=" << mu << ")\n";
  }
}

void doseprclass::put_logit(ostream& fout, option& opt, const varclass& var) const
// ロジット分析の結果を出力する
{
  fout<< "\n";
  if (opt.du() == dulog) fout << "log(Dose)";
  else fout << "Dose";
  fout << "-Logit Line:lp=ln(p/(1-p))=" << var.a() << showpos << var.b() << noshowpos << "x\n";
  fout << "  p0=1/{1+exp(" << -var.a() << ")}=" << 1.0 / (1.0 + exp(-var.a()));
  if (var.nr() > 0) fout << " (Repetitions=" << var.nr() << ")";
  fout << "\n";
  int dft = var.nd() - 1;
  if (opt.du() == dulog && var.d(0) == 0.0) --dft; // 用量0のデータがある時
  double sb = var.sxyw() * var.sxyw() / var.sxxw();
  if (opt.res("r"))
  {
    double r2 = (var.syyw() > 0.0) ? sb / var.syyw() : 0.0;
    fout << "Contribution Rate:r-squared=" << r2 << "\n";
  }
  if (opt.res("pr")) put_estlog(fout, opt, var);
  put_doseprtab(fout, "Logit Analysis Table", var.syyw(), sb, dft);
}

void doseprclass::put_estlog(ostream& fout, option& opt, const varclass& var) const
// ロジット分析による用量の逆推定値を出力する
{
  if (sgn(var.b()) == 0) return; // 逆推定できず
  double pr = opt.pr() * 100.0, p = opt.pr();
  double lp = log(p / (1.0 - p));
  double m = (lp - var.a()) / var.b();
  fout << "Dose at a Response Rate=" << pr << "%(p=" << p << "):D" << pr << "=";
  if (opt.du() == dureal) fout << "x=" << m << "\n";
  else fout << pow(10.0, m) << "(x=log(D" << pr << ")=" << m << ")\n";
  if (opt.res("cc") == false) return;
  double vr = 1.0, t = ptoz(1.0 - opt.cc());
  double g = t * t * vr / var.b() / var.b() / var.sxxw();
  if (g >= 1.0) g = 0.0;        // g>1の時は計算できないため
  double gr = 1.0 - g;
  double v = vr * (gr / var.snw() + (m - var.mxw()) * (m - var.mxw()) / var.sxxw());
  double ci = t / var.b() * sqrt(v);
  double ml = (m - g * var.mxw() - ci) / gr, mu = (m - g * var.mxw() + ci) / gr;
  fout << opt.cc() * 100.0 << "% CI of D" << pr << ":";
  if (opt.du() == dureal)
  { fout << "Lower=" << ml << "  Upper=" << mu << "\n"; }
  else
  {
    fout << "Lower=" << pow(10.0, ml) << "(x=" << ml;
    fout << ")  Upper=" << pow(10.0, mu) << "(x=" << mu << ")\n";
  }
}

void doseprclass::plot_ngp(ofstream& fout, option& opt, idd& id, varclass& var) const
// Ngraph用データ出力
{
  var.caldpline(opt);
  bool delml = false;
  if (opt.ngp("pro"))           // 反応率
  {
    fout << "-s1 -dmark,00 -o300 -zy0,1,0.2";
    if (opt.du() == dulog) fout << " -exlog";
    fout << "\n";
    plot_ngpid(fout, id);
    plot_ngppro(fout, var);
    delml = true;
  }
  if (opt.ngp("drc"))           // プロビット分析の用量反応曲線
  {
    if (delml)                   // 区切り行を出力する
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_ngpdrc(fout, opt, var);
  }
  if (opt.ngp("drl"))           // 傾向分析の用量反応直線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
      delml = true;
    }
    var.calcaline(opt);
    plot_ngpid(fout, id);
    plot_ngpdrl(fout, opt, var);
  }
  if (opt.ngp("drlp"))          // ロジット分析の用量反応曲線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
    }
    var.callpline(opt);
    plot_ngpid(fout, id);
    plot_ngpdrlp(fout, opt, var);
  }
}

void doseprclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "%DOSE ITEM(ROW)    :" << id.item(i) << "\n";
      break;
    default : fout << "%RESPONSE ITEM(COL):" << id.item(i) << "\n";
    }
  }
}

void doseprclass::plot_ngppro(ofstream& fout, const varclass& var) const
// Ngraph用に反応率を出力する
{
  fout << "%<RESPONSE RATE>\n";
  for (int i = 0; i < var.nd(); ++i)
  { fout << var.d(i) << " " << var.p(i) << "\n"; }
}

void doseprclass::plot_ngpdrc(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用にプロビット分析の用量-反応率曲線を出力する
{
  fout << "%<DOSE-RESPONSE RATE CURVE OF PROBIT ANALYSIS>\n";
  constexpr uint RDdef = 50;    // 用量の分割数
  double xmin = var.x(0);
  double xmax = var.x(var.nd() - 1);
  double range = xmax - xmin;
  double xstart = xmin - range * 0.1, xend = xmax + range * 0.1;
  double dx = (xend - xstart) / (double)RDdef;
  double x = xstart;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double y = var.a() + var.b() * x;
    double dose = (opt.du() == dulog) ? pow(10.0, x) : x;
    double pro = nedp(y - 5.0);
    fout << dose << " " << pro << "\n";
  }
}

void doseprclass::plot_ngpdrl(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に傾向分析の用量-反応率直線を出力する
{
  fout << "%<DOSE-RESPONSE RATE LINE OF TREND ANALYSIS>\n";
  double xmin = var.x(0);
  double xmax = var.x(var.nd() - 1);
  double range = xmax - xmin;
  double x = xmin - range * 0.1;
  double dose = (opt.du() == dulog) ? pow(10.0, x) : x;
  double pro = var.a() + var.b() * x;
  fout << dose << " " << pro << "\n";
  x = xmax + range * 0.1;
  dose = (opt.du() == dulog) ? pow(10.0, x) : x;
  pro = var.a() + var.b() * x;
  fout << dose << " " << pro << endl;
}

void doseprclass::plot_ngpdrlp(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用にロジット分析の用量-反応率曲線を出力する
{
  fout << "%<DOSE-RESPONSE RATE CURVE OF LOGIT ANALYSIS";
  if (opt.log() == logcomp) fout << " (repetitions=" << var.nr() << ")";
  fout << ">\n";
  constexpr uint RDdef = 50;    // 用量の分割数
  double xmin = var.x(0);
  double xmax = var.x(var.nd() - 1);
  double range = xmax - xmin;
  double xstart = xmin - range * 0.1, xend = xmax + range * 0.1;
  double dx = (xend - xstart) / (double)RDdef;
  double x = xstart;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double dose = (opt.du() == dulog) ? pow(10.0, x) : x;
    double pro = 1.0 / (1.0 + exp(-var.a() - var.b() * x));
    fout << dose << " " << pro << "\n";
  }
}

int doseprclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// クロス集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("Run-time error !");
  if (opt.check() != 0) throw option_error("Run-time option error !");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void doseprclass::put(option& opt, listidd& idlst, listftab& ftablst, int fno) const
// 結果を出力する
{
  listidd::iterator iitr = idlst.begin();
  for (listftab::iterator fitr = ftablst.begin(); fitr != ftablst.end(); ++iitr, ++fitr, ++fno)
  {
    fitr->trm();                // 行・列の合計が0のものを削除する
    if (fitr->nr() > 0)
    {
      if (fitr->nr() > 1 && fitr->nc() == 2)
      {
        varclass var(fitr->nr()); // 変数オプジェクト
        if (var.cal(opt, *fitr) > 1)
        {
          if (opt.res("res"))
          {
            put_resid(cout, *iitr);
            put_dosepr(cout, opt, *fitr, var);
            cout << "\n" << endl;
          }
          if (opt.ngpname(fno) != "")
          {
            ofstream foutngp;
            if (ofopen(foutngp, (opt.ngpname(fno)).c_str()) == false) return; // ファイルオープン失敗
            var.cal(opt, *fitr); // 期待プロビットを初期値に戻すため
            plot_ngp(foutngp, opt, *iitr, var);
            foutngp.close();
          }
        }
      }
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  doseprclass dosepr;           // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int nftab = 0, fno = 0;     // 読み込んだ度数表の数とファイル番号
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          nftab = dosepr.get(nullptr, opt, idlst, ftablst);
          dosepr.put(opt, idlst, ftablst, fno);
          fno += nftab;
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        nftab = dosepr.get(argv[i], opt, idlst, ftablst);  fok = false;
        dosepr.put(opt, idlst, ftablst, fno);
        fno += nftab;
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
