// doserce.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 15:15:43 tac>
//
//  最初の項目を用量項目、次の項目を反応項目として
//  微分ロジット分析による用量反応解析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W doserce doserce.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "doserce  Version 7.41e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:doserce [<OPTION> ...]  [<DATA FILE> ...]",
  "  The dose-response analysis by DLA(differential logit analysis) with",
  "  the fist item as dose and second item as response.",
  "<OPTION>",
  "  -dnsname[=<FILE>]:specify the file to outout DANS data",
  "    default:no output  -dnsname:-dnsname=doserce_dns.d",
  "  -du={l|r}:select the unit of dose",
  "    l:log(dose)(default)  r:dose",
  "  -f=<FILE>:reads options from the specified file",
  "  -g[={-|+}]:specify whether to use the Fieller's formula(-g:-g=+)",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={drc|drl|mlyse|myse}[,...]:select the data for Ngraph",
  "    drc:dose-response curve  drl:dose-logit line",
  "    mlyse:logit mean +/- SE by doses  myse:response mean +/- SE by doses",
  "    default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=doserce_ngp.dat",
  "  -res=[-][{def|all|cc[<C-COEF>]|reg|statly|staty|y[{mly|my|<RES>}]}]",
  "      [,...]:select output results",
  "    -:wiht -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    cc[<C-COEF>]:confidence interval(0=<C-COEF<1 cc:cc0.95)",
  "    statly:basic statistics of logit  staty:basic statistics of response",
  "    y[{mly|my|<RES>}]:inverse estimation dose for response",
  "      mly:overall logit mean  my:overall response mean",
  "      <RES>:response",
  "  -ym=<a|d|MIN>:specify the minumum value of response(-ym:-ym=a)",
  "    a:automatic setup(default)  d:setup by DLA"
};

const string Title = "=== Dose-Response Analysis by DLA ==="; // 表題
const string DNSNAMEdef = "doserce_dns.d"; // 省略時DANSファイル名
const string NGPNAMEdef = "doserce_ngp.dat"; // 省略時Ngraphファイル名

using mapds = map<double, statv<double>>; // 用量別統計量マップの別名定義
using mapds_itr = mapds::iterator;        // 反復子の別名定義
using mapds_citr = mapds::const_iterator; // const反復子の別名定義

enum dutype                     // 用量単位の種類
{
  dupnon,                       // 単位なし
  dulog,                        // 対数
  dureal,                       // 実数
  duetc                         // その他
};

enum stattype                   // 基礎統計量の種類
{
  statnon,                      // 出力なし
  statall,                      // 全て出力
  statly,                       // ロジット
  staty,                        // 反応実測値
  statetc                       // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  string vdnsname;              // DANSファイル名
  dutype vdu;                   // 用量の単位
  bool vg;                      // Fiellerの式使用フラッグ
  string vngpname;              // Ngraphファイル名
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  stattype vstat;               // 基礎統計量出力フラッグ
  double vy;                    // 用量を推定する反応値
  double vym;                   // 反応最小値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  const string& dnsname() const // DANSデータファイル名を返す
  { return vdnsname; }
  dutype du() const             // 用量の単位を返す
  { return vdu; }
  bool g() const                // Fiellerの式使用フラッグをす
  { return vg; }
  const string& ngpname() const // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  stattype stat() const         // 出力する基礎統計量の種類を返す
  { return vstat; }
  double y() const              // 用量を推定する反応値を返す
  { return vy; }
  double ym() const             // 反応最小値を返す
  { return vym; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdnsname(), vdu(dulog), vg(true), vngpname(), vngplst(), vngpmap(), vreslst(), vresmap(), vstat(staty), vy(BLANKf), vym(BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "dnsname")      // DANSファイル名指定
  {
    if (opt[1].empty()) vdnsname = DNSNAMEdef;
    else vdnsname = opt[1];
    if (vdnsname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "du")      // 用量の単位指定
  {
    if (opt[1].empty() || opt[1] == "l") vdu = dulog;
    else if (opt[1] == "r") vdu = dureal;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "g")       // Fiellerの式使用指定
  {
    if (opt[1].empty() || opt[1] == "+") vg = true;
    else if (opt[1] == "-") vg = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
    if (vngpname.size() < 3) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "stat")    // 基礎統計量出力指定
  {
    if (opt[1].empty() || opt[1] == "y") vstat = staty;
    else if (opt[1] == "-") vstat = statnon;
    else if (opt[1] == "all") vstat = statall;
    else if (opt[1] == "ly") vstat = statly;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "y")       // 用量を推定する反応値の指定
  {
    double fdata = 0.0;
    if (opt[1].empty() || opt[1] == "my") vy = BLANKf; // 反応の総平均値
    else if (opt[1] == "mly") vy = DBL_MAX;            // ロジットの総平均値
    else if (opt[1] == "-") vy = -DBL_MAX;             // 推定せず
    else vy = fdata = strtodf(opt[1]);
    if (fdata == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ym")      // 反応最小値指定
  {
    double fdata = 0.0;
    if (opt[1].empty() || opt[1] == "a") vym = BLANKf; // 自動設定(省略時と同じ)
    else if (opt[1] == "d") vym = -DBL_MAX; // 2次回帰分析によって設定
    else vym = fdata = strtodf(opt[1]);     // 最小値指定
    if (fdata == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)   // オプションのチェックとセット
{
  if (ni != 2) return disperr(vpname, "Number of items must be 2 !");
  if (set_ngp() < 0) return -2; // -resオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["drc"] = vngpmap["drl"] = vngpmap["mlyse"] = vngpmap["myse"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["drc"] = vngpmap["drl"] = vngpmap["mlyse"] = vngpmap["myse"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["staty"] = vresmap["y"] = true;
  vresmap["statly"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["statly"] = vresmap["staty"] = vresmap["y"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["staty"] = vresmap["y"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["statly"] = vresmap["staty"] = vresmap["y"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "ymly")
      { vresmap["y"] = true;  vy = DBL_MAX; }
      else if (res == "ymy")
      { vresmap["y"] = true;  vy = BLANKf; }
      else if (res[0] == 'y' && res.size() > 1)
      {
        vy = strtodf(res.substr(1)); // 反応値
        if (vy == BLANKf) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["y"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  int vnd;                      // 用量数
  mapds vdsmap;                 // 用量別反応統計量マップ
  mapds vdslymap;               // 用量別ロジット統計量マップ
  pairdbl drange;               // 用量の最小値と最大値
  pairdbl yrange;               // 反応の最小値と最大値
  statcor<double> statcdy;      // 用量と反応の相関用基礎統計量
  statcor<double> statcdly;     // 用量とロジットの相関用基礎統計量
  double vsd;                   // 用量平方和
  double vym;                   // y最小値
  double vy0;                   // y定数
  double vc;                    // 回帰係数
  double vm;                    // y最大値
  int vnr;                      // 繰り返し計算数
  int init_drc(const option& opt); // パラメータの初期値を設定する
  int cal_smat(double& ym, mat<double>& smat, const option& opt) const;
  // 2次回帰分析用に積和行列を計算する
  double cal_y0(double ym, double c, double m, const option& opt) const;
  // y0をロジット分析によって求める
  int cal_drc(const option& opt, ldata datalst);
  // ロジット分析によってパラメータを推定する
  int cal_reglog(const option& opt, ldata datalst);
  // ロジット分析によってパラメータを推定する(ymとMは更新しない)
  int cal_replog(const option& opt, ldata datalst);
  // 繰り返しロジット分析によってパラメータを更新する
  int cal_dslymap(const option& opt, ldata datalst);
  // 最終的なロジット分析を行って用量別ロジットマップをセットする
public:
  varclass() : vnd(0), vdsmap(), vdslymap(), drange(DBL_MAX, -DBL_MAX), yrange(DBL_MAX, -DBL_MAX), statcdy(), statcdly(), vsd(0.0), vym(0.0), vy0(0.0), vc(0.0), vm(0.0), vnr(0) // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int nd() const                // 用量数を返す
  { return vnd; }
  const mapds& dsmap() const    // 用量別統計量マップを返す
  { return vdsmap; }
  const mapds& dslymap() const  // 用量別ロジット統計量マップを返す
  { return vdslymap; }
  int dyn() const               // 用量反応の例数を返す
  { return statcdy.n(); }
  double dymx() const           // 用量反応の用量平均値を返す
  { return statcdy.mx(); }
  double dymy() const           // 用量反応の反応平均値を返す
  { return statcdy.my(); }
  double dysxx() const          // 用量反応の用量平方和を返す
  { return statcdy.sxx(); }
  double dysyy() const          // 用量反応の反応平方和を返す
  { return statcdy.syy(); }
  double dysdy() const          // 用量反応の反応標準偏差を返す
  { return statcdy.sdy(); }
  double dysey() const          // 用量反応の反応標準誤差を返す
  { return statcdy.sey(); }
  double dysxy() const          // 用量反応の用量と反応積和を返す
  { return statcdy.sxy(); }
  int dlyn() const              // 用量ロジットの例数を返す
  { return statcdly.n(); }
  double dlymx() const          // 用量ロジットの用量平均値を返す
  { return statcdly.mx(); }
  double dlymy() const          // 用量ロジットのロジット平均値を返す
  { return statcdly.my(); }
  double dlysxx() const         // 用量ロジットの用量平方和を返す
  { return statcdly.sxx(); }
  double dlysyy() const         // 用量ロジットのロジット平方和を返す
  { return statcdly.syy(); }
  double dlysdy() const         // 用量ロジットのロジット標準偏差を返す
  { return statcdly.sdy(); }
  double dlysey() const         // 用量ロジットのロジット標準誤差を返す
  { return statcdly.sey(); }
  double dlysxy() const         // 用量ロジットの用量とロジット積和を返す
  { return statcdly.sxy(); }
  double dmin() const           // 用量の最小値を返す
  { return drange.first; }
  double dmax() const           // 用量の最大値を返す
  { return drange.second; }
  double ymin() const           // 反応の最小値を返す
  { return yrange.first; }
  double ymax() const           // 反応の最大値を返す
  { return yrange.second; }
  double sd() const             // 用量平方和を返す
  { return vsd; }
  double ym() const             // y最小値を返す
  { return vym; }
  double y0() const             // y定数を返す
  { return vy0; }
  double c() const              // 回帰係数を返す
  { return vc; }
  double m() const              // y最大値を返す
  { return vm; }
  int nr() const                // 繰り返し計算数を返す
  { return vnr; }
  int cal(const option& opt, ldata datalst);
  // データを読み込みロジット分析を行う
};

int varclass::init_drc(const option& opt) // パラメータの初期値を設定する
{
  vy0 = (yrange.first > 0.0) ? 0.5 * yrange.first : 0.1;
  vm = (yrange.second > vy0) ? yrange.second : vy0 * 2.0;
  double b1 = (statcdy.sxx() > 0.0) ? statcdy.sxy() / statcdy.sxx() : 0.0;
  vc = (sgn(vm) > 0) ? 8.0 * b1 / vm : 0.0; // 初期値を回帰直線によって設定
  vym = opt.ym();
  int p = 2;
  if (vym == -DBL_MAX) p = 3;   // ymを2次回帰分析によって推測する
  int dim = p + 1;
  int dyidx = p;                // dyのインデックス
  mat<double> smat(dim);          // 積和行列
  int nd = cal_smat(vym, smat, opt);
  if (nd < p) return nd;        // 2次回帰分析不可能
  double b0 = 0.0, b2 = 0.0, ym0 = vym, y00 = vy0, c0 = vc, m0 = vm;
  if (vym == -DBL_MAX)          // ymを2次回帰分析から求める
  {
    b0 = smat(0, dyidx);  b1 = smat(1, dyidx);  b2 = smat(2, dyidx);
    double sqrtb = sqrt(b1 * b1 - 4.0 * b2 * b0);
    ym0 = (-b1 + sqrtb) / (2.0 * b2);
    double ymax = (-b1 - sqrtb) / (2.0 * b2);
    m0 = ymax - ym0;
    c0 = -b2 * m0;
  }
  else                          // ymをデータから求める／-ymで指定
  {
    b1 = smat(0, dyidx);  b2 = smat(1, dyidx);
    c0 = b1;  m0 = -b1 / b2;  ym0 = vym;
  }
  if (sgn(c0) < 0) return -1;   // 2次回帰分析結果不適切
  y00 = cal_y0(ym0, c0, m0, opt); // y0をロジット分析によって求める
  if (y00 == BLANKf) return -2;
  vym = ym0;  vy0 = y00;  vc = c0;  vm = m0;
  return nd;                    // 有効な用量数返す
}

int varclass::cal_smat(double& ym, mat<double>& smat, const option& opt) const
// 2次回帰分析用に積和行列を計算する
{
  int dim = smat.r();           // 行列の次元
  int p = dim - 1;              // 2次回帰式のパラメータ数
  vecdbl fdatavec(dim);
  size_t offset = 0;
  if (dim == 4) { fdatavec[0] = 1.0;  offset = 1; } // 最小値も推定する時
  size_t dyidx = p;             // dyのインデックス
  double md1 = 0.0, md2 = 0.0, y1 = 0.0, y2 = 0.0;
  int nd = 0;
  for (mapds_citr dsitr = vdsmap.begin(); dsitr != vdsmap.end(); ++dsitr)
  {
    double dose = dsitr->first, y = dsitr->second.mean();
    if (ym == BLANKf && opt.du() == dulog && sgn(dose) == 0) ym = y;
    if (opt.du() == dureal || dose > 0.0)
    {
      if (ym == BLANKf) ym = 0.0; // dose=0が無い時
      double mdose = (opt.du() == dulog) ? log10(dose) : dose;
      if (nd < 1)
      { md1 = mdose;  y1 = y;  nd = 1; }
      else if (nd == 1)
      { md2 = mdose;  y2 = y;  nd = 2; }
      else
      {
        double mdd1 = md2 - md1, mdd2 = mdose - md2;
        double eps = DBL_EPSILON * fmax(5.0, fabs(mdd1)); // 計算誤差
        double w1 = (fabs(mdd1) < eps) ? 0.0 : 1.0 / mdd1;
        double w2 = (fabs(mdd2) < eps) ? 0.0 : 1.0 / mdd2;
        double b1 = (y2 - y1) * w1, b2 = (y - y2) * w2;
        w1 = fabs(w1);  w2 = fabs(w2);
        double dy = (fabs(w1 + w2) < eps) ? 0.0 : (b1 * w1 + b2 * w2) / (w1 + w2);
        double my = y2;
        if (ym != -DBL_MAX) my -= ym; // 最小値を引く
        double fd = my;               // 例数で重み付けしない
        for (int i = 0; i < 2; ++i)
        {
          fdatavec[offset + i] = fd;
          fd *= my;
        }
        fdatavec[dyidx] = dy;
        for (int i = 0; i < dim; ++i)
        {
          fd = fdatavec[i];
          for (int j = i; j < dim; ++j) smat.add(i, j, fd * fdatavec[j]);
        }
        md1 = md2;  y1 = y2;
        md2 = mdose;  y2 = y;  ++nd;
      }
    }
  }
  if (nd - 2 < p) return nd;    // 2次回帰分析不可能
  for (size_t i = 0; i < dyidx; ++i)
  {
    for (size_t j = i + 1; j <= dyidx; ++j) smat.assign(j, i, smat(i, j));
  }
  for (size_t i = 0; i < dyidx; ++i) smat.sweep(i);
  return nd;                    // 有効な用量数を返す
}

double varclass::cal_y0(double ym, double c, double m, const option& opt) const
// y0をロジット分析によって求める
{
  statcor<double> statcdly0;
  for (mapds_citr dsitr = vdsmap.begin(); dsitr != vdsmap.end(); ++dsitr)
  {
    double dose = dsitr->first, my = dsitr->second.mean() - ym;
    if (opt.du() == dureal || sgn(dose) > 0)
    {
      double mdose = (opt.du() == dulog) ? log10(dose) : dose;
      if (0.0 < my && my < m)  // myが範囲内の時だけ集計
      {
        double ly = log(my / (m - my));
        statcdly0.sigma(mdose, ly); // 例数で重み付けしない
      }
    }
  }
  if (statcdly0.n() < 1) return BLANKf; // ロジットがない時
  double ly0 = statcdly0.my() - c * statcdly0.mx();
  double y0 = ym + m / (1.0 + exp(-ly0));
  if (y0 - ym >= m) return BLANKf; // 不適切なy0
  return y0;                  // y0を返す
}

int varclass::cal_drc(const option& opt, ldata datalst)
// ロジット分析によってパラメータを推定する
{
  int ncomp = cal_reglog(opt, datalst); // 仮パラメータによるロジット分析
  if (ncomp < 0) return -1;             // ロジット分析不可能
  vnr = 0;
  if (ncomp > 0) vnr = cal_replog(opt, datalst);
  if (vnr < 0) return -2;                 // 繰り返しロジット分析発散
  int nd = cal_dslymap(opt, datalst); // 最終的なロジット分析と用量別マップ作成
  if (nd < 2) return -3;        // ロジット分析不可能
  vsd = 0.0;
  for (mapds_citr dsitr = vdslymap.begin(); dsitr != vdslymap.end(); ++dsitr)
  {
    double ly = dsitr->second.mean();
    vsd += (double)(dsitr->second.n()) * ly * ly;
  }
  vsd -= (double)(statcdly.n()) * statcdly.my() * statcdly.my();
  return vnr;                   // 繰り返し数を返す
}

int varclass::cal_reglog(const option& opt, ldata datalst)
// ロジット分析によってパラメータを推定する(ymとMは更新しない)
{
  double y = vy0 - vym;
  double ly0 = log(y / (vm - y)); // ロジット回帰式の定数
  statcdly.clear();              // statcxlyを繰り返し使用するので初期化する
  int ncomp = 0;
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double dose = ditr->fdata(0), yo = ditr->fdata(1);
    if (dose != BLANKf && yo != BLANKf)
    {
      if (opt.du() == dureal || sgn(dose) > 0)
      {
        double mdose = (opt.du() == dulog) ? log10(dose) : dose;
        double y = yo - vym;
        double ly = ly0 + vc * mdose; // ロジット推定値(外れ値補完用)
        if (0.0 < y && y < vm) ly = log(y / (vm - y)); // yが範囲内
        else ++ncomp;
        statcdly.sigma(mdose, ly);
      }
    }
  }
  if (statcdly.n() < 2) return -1; // ロジット分析不可能
  double mx = statcdly.mx(), my = statcdly.my();
  double sxx = statcdly.sxx(), sxy = statcdly.sxy();
  vc = (sgn(sxx) > 0) ? sxy / sxx : 0.0;
  ly0 = my - vc * mx;
  vy0 = vym + vm / (1.0 + exp(-ly0));
  if (sgn(vc) < 0 || vy0 - vym >= vm) return -2; // ロジット分析結果不適切
  return ncomp;                 // 外れ値の補完数を返す
}

int varclass::cal_replog(const option& opt, ldata datalst)
// 繰り返しロジット分析によってパラメータを更新する
{
  constexpr int Rmax = 100;     // 繰り返し数最大値
  double syypre = statcdly.syy();
  const double Seps = 1.0e-5 * fmax(5.0, syypre);
  int nr = 0, ncomp = 0;
  bool comp = true;
  while (comp)
  {
    ncomp = cal_reglog(opt, datalst);
    double syy = statcdly.syy();  ++nr;
    if (ncomp < 0 || fabs(syy - syypre) < Seps || nr >= Rmax) comp = false;
    else syypre = syy;
  }
  if (ncomp < 0) return ncomp;  // 繰り返し計算発散
  return nr;                    // 繰り返し数を返す
}

int varclass::cal_dslymap(const option& opt, ldata datalst)
// 最終的なロジット分析を行って用量別ロジットマップをセットする
{
  double y = vy0 - vym;
  double ly0 = log(y / (vm - y)); // ロジット回帰式の定数
  statcdly.clear();
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double dose = ditr->fdata(0), yo = ditr->fdata(1);
    if (dose != BLANKf && yo != BLANKf)
    {
      if (opt.du() == dureal || sgn(dose) > 0)
      {
        double mdose = (opt.du() == dulog) ? log10(dose) : dose;
        double y = yo - vym;
        double ly = ly0 + vc * mdose; // ロジット推定値(外れ値はこれで補完)
        if (0.0 < y && y < vm) ly = log(y / (vm - y)); // myが範囲内
        statcdly.sigma(mdose, ly);
        mapds_itr dslyitr = vdslymap.find(dose);
        if (dslyitr == vdslymap.end()) // 新規用量
        {
          statv<double> statvi(ly);
          vdslymap.insert(make_pair(dose, statvi));
        }
        else                      // すでにある用量
        { dslyitr->second.sigma(ly);  }
      }
    }
  }
  double mx = statcdly.mx(), my = statcdly.my();
  double sxx = statcdly.sxx(), sxy = statcdly.sxy();
  vc = (sgn(sxx) > 0) ? sxy / sxx : 0.0;
  ly0 = my - vc * mx;
  vy0 = vym + vm / (1.0 + exp(-ly0));
  if (sgn(vc) < 0 || vy0 - vym >= vm) return -1; // ロジット分析結果不適切
  return (int)(vdslymap.size()); // マップ数を返す
}

int varclass::cal(const option& opt, ldata datalst)
// データを読み込みロジット分析を行う
{
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double dose = ditr->fdata(0), y = ditr->fdata(1);
    if (dose != BLANKf && y != BLANKf) // dose=0も集計する
    {
      mapds_itr dsitr = vdsmap.find(dose);
      if (dsitr == vdsmap.end()) // 新規用量
      {
        statv<double> statvi(y);
        vdsmap.insert(make_pair(dose, statvi));
      }
      else                      // すでにある用量
      { dsitr->second.sigma(y);  }
      if (opt.du() == dureal || sgn(dose) > 0)
      {
        if (dose < drange.first) drange.first = dose;
        if (dose > drange.second) drange.second = dose;
        if (y < yrange.first) yrange.first = y;
        if (y > yrange.second) yrange.second = y;
        double mdose = (opt.du() == dulog) ? log10(dose) : dose; // メタ用量
        statcdy.sigma(mdose, y);
      }
    }
  }
  vnd = init_drc(opt);
  if (vnd < 1) return vnd;      // 計算不可能
  vnr = cal_drc(opt, datalst);
  return vnd;                   // 用量数を返す
}

class dosercclass               // dosercメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_mstat(ostream& fout, option& opt, const varclass& var) const;
  // 用量別基礎統計量を出力する
  void put_drc(ostream& fout, const option& opt, const varclass& var) const;
  // 用量反応解析結果を出力する
  void put_estd(ostream& fout, option& opt, const varclass& var) const;
  // 用量を逆推定する
  void put_drtab(ostream& fout, const varclass& var) const;
  // 用量反応用分散分析表を出力する
  void put_dns(const option& opt, idd& id, ldata& datalst, const varclass& var) const;
  // xとyとy推測値とlyとly推測値をデータファイル形式で出力する
  void put_dnsid(ofstream& fout, idd& id) const;
  // DANSデータのタイトルとデータ識別情報を出力する
  void plot_ngp(option& opt, idd& id, const varclass& var) const;
  // Ngraph用にデータを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpdrc(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に用量反応曲線を出力する
  void plot_ngpdrl(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に用量ロジット直線を出力する
  void plot_ngpmyse(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に反応の平均値±SEを出力する
  void plot_ngpmlyse(ofstream& fout, const varclass& var) const;
  // Ngraph用にロジットの平均値±SEを出力する
public:
  dosercclass() : fname()       // コンストラクタ
  {}
  ~dosercclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void dosercclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "DOSE ITEM    :" << id.item(0) << "\n";
  fout << "RESPONSE ITEM:" << id.item(1) << endl;
}

void dosercclass::put_mstat(ostream& fout, option& opt, const varclass& var) const
// 用量別基礎統計量を出力する
{
  constexpr int LENline = 70;
  fout << "\n    Dose     ";
  if (opt.du() == dulog) fout << "x=log(Dose)";
  else fout << "     x=Dose";
  fout << "         n        mean          SD          SE\n";
  putline(fout, LENline);
  bool putstat = false;
  if (opt.res("staty"))
  {
    for (mapds_citr dsitr = (var.dsmap()).begin(); dsitr != (var.dsmap()).end(); ++dsitr)
    {
      fout << setw(8) << dsitr->first;
      if (opt.du() == dulog)
      {
        if (sgn(dsitr->first) > 0) fout << setw(16) << log10(dsitr->first);
        else fout << "     (exclusion)";
      }
      else
      { fout << setw(16) << dsitr->first; }
      fout << setw(10) << dsitr->second.n() << setw(12) << dsitr->second.mean();
      fout << setw(12) << dsitr->second.sd() << setw(12) << dsitr->second.se();
      fout << "\n";
    }
    putline(fout, LENline);
    string tstr("TOTAL");
    int npad = tstr.size() - strclen(tstr, CHARwide);
    fout << setw(8 + npad) << tstr << setw(16) << var.dymx();
    fout << setw(10) << var.dyn() << setw(12) << var.dymy();
    fout << setw(12) << var.dysdy() << setw(12) << var.dysey() << "\n";
    putstat = true;
  }
  if (opt.res("statly"))
  {
    if (putstat) putline(fout, LENline);
    fout << "LOGIT\n";
    for (mapds_citr dsitr = (var.dslymap()).begin(); dsitr != (var.dslymap()).end(); ++dsitr)
    {
      fout << setw(8) << dsitr->first;
      if (opt.du() == dulog) fout << setw(16) << log10(dsitr->first);
      else fout << setw(16) << dsitr->first;
      fout << setw(10) << dsitr->second.n() << setw(12) << dsitr->second.mean();
      fout << setw(12) << dsitr->second.sd() << setw(12) << dsitr->second.se();
      fout << "\n";
    }
    putline(fout, LENline);
    string tstr("TOTAL");
    int npad = tstr.size() - strclen(tstr, CHARwide);
    fout << setw(8 + npad) << tstr << setw(16) << var.dlymx();
    fout << setw(10) << var.dlyn() << setw(12) << var.dlymy();
    fout << setw(12) << var.dlysdy() << setw(12) << var.dlysey() << "\n";
  }
}

void dosercclass::put_drc(ostream& fout, const option& opt, const varclass& var) const
// 用量反応解析結果を出力する
{
  double y = var.y0() - var.ym();
  double ly0 = log(y / (var.m() - y));
  double sxx = var.dlysxx(), sxy = var.dlysxy();
  double sb = (sgn(sxx) > 0) ? sxy * sxy / sxx : 0.0;
  double r2 = (sgn(var.sd()) > 0) ? sb / var.sd() : 0.0;
  fout << "\n";
  if (opt.du() == dulog) fout << "log(Dose)";
  else fout << "Dose";
  fout << "-Logit Line:ly=ly0+cx=" << ly0 << "+" << var.c() << "x\n";
  fout << "Logit:ly=ln[(y-ym)/{M-(y-ym)}]  M=" << var.m() << "  ym=" << var.ym();
  fout << "\nIntercept:ly0=ln[(y0-ym)/{M-(y0-ym)}]=" << ly0 << "  y0=" << var.y0() << "\n";
  fout << "Dose-Response Curve:y=ym+M/{1+exp(-ly0-cx)}  Repetitions=";
  if (var.nr() > 0) fout << var.nr();
  else if (var.nr() < 0) fout << "DIVERGED !";
  fout << "\nMin:ym=" << var.ym() << "  Max:ym+M=" << var.ym() << "+" << var.m() << "=" << var.ym() + var.m();
  fout << "\nMultiple Contribution Rate:R-squared=" << r2 << "  R=" << sqrt(r2) << endl;
}

void dosercclass::put_estd(ostream& fout, option& opt, const varclass& var) const
// 用量を逆推定する
{
  double ye = (opt.y() == BLANKf) ? var.dymy() : opt.y();
  double y = ye - var.ym();
  double ly = (y != DBL_MAX && y > 0.0) ? log(y / (var.m() - y)) : 0.0;
  if (opt.y() == DBL_MAX)
  {
    ly = var.dlymy();
    y = var.m() / (1.0 + exp(-ly));
    ye = y + var.ym();
  }
  fout << "\nDose at Response(y)=" << ye << "(ly=" << "" << ly << "):";
  if (y < 0.0 || var.m() <= y || sgn(var.c()) == 0)
  {
    fout << "INVERSE ESTIMATAION IS IMPOSSIBLE !\n";
    return;
  }
  y = var.y0() - var.ym();
  double ly0 = log(y / (var.m() - y));
  double x = (ly - ly0) / var.c();
  if (opt.du() == dureal) fout << "D=x=" << x;
  else fout << "D=10^(" << x << ")=" << pow(10.0, x);
  fout << "\n";
  if (opt.res("cc") == false) return;
  int nt = var.dlyn();
  int dft = nt - 1, dfd = var.nd() - 1, dfb = 1;
  int dfl = dfd - dfb, dfr = dft - dfd;
  double st = var.dlysyy();
  const double Seps = DBL_EPSILON * fmax(5.0, st);
  double sr = st - var.sd();
  if (dfr < 1 || sr < Seps) sr = 0.0;
  double sxx = var.dlysxx(), sxy = var.dlysxy();
  double sb = (sgn(sxx) > 0) ? sxy * sxy / sxx : 0.0;
  double sl = var.sd() - sb;
  if (dfl < 1 || sl < Seps) sl = 0.0;
  if (dfd == dft) { sr = sl;  dfr = dfl; } // 用量群が1例だけの時
  double vr = (dfr > 0) ? sr / (double)dfr : 0.0;
  double mx = var.dlymx(), my = var.dlymy();
  double cf = (ly - my) / var.c();
  double t = ptot(1.0 - opt.cc(), dfr);
  double g = t / var.c() * t / var.c() * vr / sxx;
  if (opt.g() == false || g >= 1.0) g = 0.0; // g>1だと計算できないため
  double gr = 1.0 - g;
  double v = vr * (gr * (1.0 + 1.0 / (double)nt) + cf * cf / sxx);
  double ci = t / var.c() * sqrt(v);
  double xlow = (x - g * mx - ci) / gr, xup = (x - g * mx + ci) / gr;
  fout << opt.cc() * 100.0 << "%CI of Dose:";
  if (opt.du() == dureal) fout << "Lower=" << xlow << "  Upper=" << xup;
  else fout << "Lower=10^(" << xlow << ")=" << pow(10.0, xlow) << "  Upper=10^(" << xup << ")=" << pow(10.0, xup);
  fout << endl;
}

void dosercclass::put_drtab(ostream& fout, const varclass& var) const
// 用量反応用分散分析表を出力する
{
  constexpr int LENline = 74;
  int dft = var.dlyn() - 1, dfd = var.nd() - 1, dfb = 1;
  int dfl = dfd - dfb, dfr = dft - dfd;
  double st = var.dlysyy();
  const double Seps = DBL_EPSILON * fmax(5.0, st);
  double sr = st - var.sd();
  if (dfr < 1 || sr < Seps) sr = 0.0;
  double sxx = var.dlysxx(), sxy = var.dlysxy();
  double sb = (sgn(sxx) > 0) ? sxy * sxy / sxx : 0.0;
  double sl = var.sd() - sb;
  if (dfl < 1 || sl < Seps) sl = 0.0;
  if (dfd == dft) { sr = sl;  dfr = dfl; } // 用量群が1例だけの時
  string tstr("ANOVA Table of Logit");
  int nstr = tstr.size();
  fout << "\n" << setw((LENline - nstr) / 2 + nstr) << tstr;
  fout << "\nFACTOR            SS        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  calvfp(v, f, pr, sb, dfb, sr, dfr);
  fout << "REGRESSION" << setw(10) << sb << setw(10) << dfb;
  fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, sl, dfl, sr, dfr);
  fout << "LOF" << setw(17) << sl << setw(10) << dfl << setw(14) << v;
  if (dft > dfd) fout << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sd(), dfd, sr, dfr);
  fout << "DOSE" << setw(16) << var.sd() << setw(10) << dfd;
  if (dft > dfd)
  {
    fout << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
    v = (dfr > 0) ? sr / (double)dfr : 0.0;
    fout << "RESIDUALS" << setw(11) << sr << setw(10) << dfr;
    fout << setw(14) << v << "\n";
    putline(fout, LENline);
    fout << "TOTAL" << setw(15) << st << setw(10) << dft;
  }
  fout << endl;
}

void dosercclass::put_dns(const option& opt, idd& id, ldata& datalst, const varclass& var) const
// xとyとy推測値とlyとly推測値をデータファイル形式で出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.dnsname()).c_str()) == false) return; // ファイルオープン失敗
  put_dnsid(fout, id);
  datac tdata(6);
  double y = var.y0() - var.ym();
  double ly0 = log(y / (var.m() - y));
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double dose = ditr->fdata(0), yo = ditr->fdata(1);
    tdata.assign(0, dose);  tdata.assign(1, yo);
    if (dose == BLANKf || yo == BLANKf || (opt.du() == dulog && sgn(dose) <= 0))
    {
      tdata.assign(2, (dose != BLANKf && sgn(dose) > 0) ? log(dose) : BLANKf);
      tdata.assign(3, BLANKs);  tdata.assign(4, BLANKs);
      tdata.assign(5, BLANKs);
    }
    else
    {
      tdata.assign(2, (sgn(dose) > 0) ? log(dose) : BLANKf);
      double mdose = (opt.du() == dulog) ? log10(dose) : dose;
      double lye = ly0 + var.c() * mdose; // ロジット推測値
      double ye = var.ym() + var.m() / (1.0 + exp(-lye)); // y推測値
      tdata.assign(3, ye);
      double ly = lye;
      y = yo - var.ym();
      if (0.0 < y && y < var.m()) ly = log(y / (var.m() - y));
      tdata.assign(4, ly);
      tdata.assign(5, lye);
    }
    tdata.put(fout);
  }
}

void dosercclass::put_dnsid(ofstream& fout, idd& id) const
// DANSデータのタイトルとデータ識別情報を出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  idd tid(id.title(), 6);
  tid.assign_item(0, id.item(0));
  tid.assign_item(1, id.item(1));
  ostringstream ostr;
  ostr << "log(" << id.item(0) << ")";
  tid.assign_item(2, ostr.str());
  ostr.str("");
  ostr << "Estimates of " << id.item(1);
  tid.assign_item(3, ostr.str());
  ostr.str("");
  ostr << "Logit of " << id.item(1);
  tid.assign_item(4, ostr.str());
  tid.assign_item(5, "Estimates of the Logit");
  tid.put_item(fout, formtag);
}

void dosercclass::plot_ngp(option& opt, idd& id, const varclass& var) const
// Ngraph用にデータを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("myse"))          // 反応の用量別平均値±SE
  {
    fout << "-s1 -dmark,00 -o300";
    if (opt.du() == dulog) fout << " -exlog";
    fout << "\n";
    plot_ngpid(fout, id);
    plot_ngpmyse(fout, opt, var);
    delml = true;
  }
  if (opt.ngp("drc"))           // 用量反応曲線
  {
    if (delml)                   // 区切り行を出力する
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    plot_ngpdrc(fout, opt, var);
  }
  if (opt.ngp("mlyse"))         // ロジットの用量別平均値±SE
  {
    if (delml)                   // 区切り行を出力する
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dmark,00 -o300";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
      delml = true; 
   }
    plot_ngpid(fout, id);
    plot_ngpmlyse(fout, var);
  }
  if (opt.ngp("drl"))           // 用量ロジット直線
  {
    if (delml)                   // 区切り行を出力する
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
    }
    plot_ngpid(fout, id);
    plot_ngpdrl(fout, opt, var);
  }
}

void dosercclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%DOSE ITEM    :" << id.item(0) << "\n";
  fout << "%RESPONSE ITEM:" << id.item(1) << endl;
}

void dosercclass::plot_ngpdrc(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に用量反応曲線を出力する
{
  fout << "%<DOSE-RESPONSE CURVE(LOGISTIC CURVE)>\n";
  constexpr int RDdef = 50;     // グラフの描画範囲分割数省略時値
  double xmin = (opt.du() == dureal) ? var.dmin() : log10(var.dmin());
  double xmax = (opt.du() == dureal) ? var.dmax() : log10(var.dmax());
  double xrange = xmax - xmin;
  double xstart = xmin - xrange * 0.1, xend = xmax + xrange * 0.1;
  double dx = (xend - xstart) / (double)RDdef;
  double x = xstart;
  double y = var.y0() - var.ym();
  double ly0 = log(y / (var.m() - y));
  for (int i = 0; i <= RDdef; ++i, x += dx)
  {
    double dose = (opt.du() == dureal) ? x : pow(10.0, x);
    fout << dose << " " << var.ym() + var.m() / (1.0 + exp(-ly0 - var.c() * x)) << endl;
  }
}

void dosercclass::plot_ngpdrl(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に用量ロジット直線を出力する
{
  fout << "%<DOSE-LOGIT LINE>\n";
  double xmin = (opt.du() == dureal) ? var.dmin() : log10(var.dmin());
  double xmax = (opt.du() == dureal) ? var.dmax() : log10(var.dmax());
  double xrange = xmax - xmin;
  double xstart = xmin - xrange * 0.1, xend = xmax + xrange * 0.1;
  double y = var.y0() - var.ym();
  double ly0 = log(y / (var.m() - y));
  double ly = ly0 + var.c() * xstart;
  double dose = (opt.du() == dulog) ? pow(10.0, xstart) : xstart;
  fout << dose << " " << ly << "\n";
  ly = ly0 + var.c() * xend;
  dose = (opt.du() == dulog) ? pow(10.0, xend) : xend;
  fout << dose << " " << ly << endl;
}

void dosercclass::plot_ngpmyse(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に反応の平均値±SEを出力する
{
  fout << "%<RESPONSE MEAN +/- SE(draw SE with -derrorbar_y -w40 -o500)>\n";
  for (mapds_citr dsitr = (var.dsmap()).begin(); dsitr != (var.dsmap()).end(); ++dsitr)
  {
    if (opt.du() == dureal || sgn(dsitr->first) > 0)
    {
      double se = dsitr->second.se();
      fout << dsitr->first << " " << dsitr->second.mean() << " " << -se << " " << se << endl;
    }
  }
}

void dosercclass::plot_ngpmlyse(ofstream& fout, const varclass& var) const
// Ngraph用に反応の平均値±SEを出力する
{
  fout << "%<LOGIT MEAN +/- SE(draw SE with -derrorbar_y -w40 -o500)>\n";
  for (mapds_citr dsitr = (var.dslymap()).begin(); dsitr != (var.dslymap()).end(); ++dsitr)
  {
    double se = dsitr->second.se();
    fout << dsitr->first << " " << dsitr->second.mean() << " " << -se << " " << se << endl;
  }
}

int dosercclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void dosercclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var;                 // 変数オブジェクト
  int nd = var.cal(opt, datalst);
  if (opt.res("res"))           // 解析結果出力
  {
    put_resid(cout, id);
    if (opt.res("statly") || opt.res("staty")) put_mstat(cout, opt, var);
    if (nd > 0)                 // 微分ロジット分析実行
    {
      if (opt.res("reg"))
      {
        put_drc(cout, opt, var);
        if (opt.res("y")) put_estd(cout, opt, var);
        put_drtab(cout, var);
      }
    }
    cout << "\n" << endl;
  }
  if (nd > 0)
  {
    if (opt.dnsname() != "") put_dns(opt, id, datalst, var);
    if (opt.ngpname() != "") plot_ngp(opt, id, var);
  }
}

int main(int argc, char *argv[]) // メインルーチン
{
  dosercclass doserc;           // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) doserc.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { doserc.get(argv[i], opt, id, datalst);  fok = false; }
    }
    doserc.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
