// dosere.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 15:51:07 tac>
//
//  最初の項目を用量項目、次の項目を反応項目とした用量反応解析を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W dosere dosere.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "dosere  Version 7.41e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:dosere [<OPTION> ...]  [<DATA FILE> ...]",
  "  The dose-response analysis with the fist item as dose and second item as",
  "  response.",
  "<OPTION>",
  "  -c=<C1>,<C2>[,<C3>,…]:specify coefficients of maximum contrast method",
  "    sum of coefficients must be 0(default:no maximum contrast method)",
  "  -du={l|r}:select the unit of dose",
  "    l:log(dose)(default)  r:dose",
  "  -f=<FILE>:reads options from the specified file",
  "  -g[={-|+}]:specify whether to use the Fieller's formula(-g:-g=+)",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={drl|mse}[,...]:select the data for Ngraph",
  "    drl:dose-response line  mse:mean +/- SE  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=dosere_ngp.dat",
  "  -res=[-][{def|all|cc[<C-COEF>]|reg|stat|y[{m|<RES>}]}][,...]",
  "      :select output results",
  "    -:wiht -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    cc[<C-COEF>]:confidence interval(0=<C-COEF<1 cc:cc0.95)",
  "    reg:regression analysis  stat:basic statistics",
  "    y[{m|<RES>}]:inverse estimation dose for response",
  "      m:overall response mean  RES:response"
};

const string Title = "=== Dose-Response Analysis ==="; // 表題
const string NGPNAMEdef = "dosere_ngp.dat"; // 省略時Ngraphファイル名

using mapds = map<double, statv<double>>; // 用量別統計量マップの別名定義
using mapds_itr = mapds::iterator;        // 指定子の別名定義
using mapds_citr = mapds::const_iterator; // const指定子の別名定義

enum dutype                     // 用量単位の種類
{
  dupnon,                       // 単位なし
  dulog,                        // 対数
  dureal,                       // 実数
  duetc                         // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  lstr clst;                    // 最大対比法用係数リスト
  vecdbl vcvec;                 // 最大対比法用係数ベクトル
  dutype vdu;                   // 用量の単位
  bool vg;                      // Fiellerの式使用フラッグ
  string vngpname;              // Ngraphファイル基本名
  string vngpnamen;             // Ngraphファイル名(番号付き)
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  double vy;                    // 用量を推定する反応値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  size_t csize() const          // 最大対比法用係数の数を返す
  { return vcvec.size(); }
  double c(size_t idx) const    // 最大対比法用係数を返す
  { return (idx < vcvec.size()) ? vcvec[idx] : 0.0; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  dutype du() const             // 用量の単位を返す
  { return vdu; }
  bool g() const                // Fiellerの式使用フラッグをす
  { return vg; }
  const string& ngpname()       // Ngraphデータファイル名を返す
  { return vngpname; }
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 分析結果出力フラッグを返す
  double y() const              // 用量を推定する反応値を返す
  { return vy; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t nd);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), clst(), vcvec(), vcc(CCdef), vdu(dulog), vg(true), vngpname(), vngpnamen(), vngplst(), vngpmap(), vreslst(), vresmap(), vy(BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngppos = vngpmap.find(code);
  if (ngppos == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "c")            // 最大対比法用係数の指定
  {
    if (strsplit(clst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "du")      // 用量の単位指定
  {
    if (opt[1].empty() || opt[1] == "l") vdu = dulog;
    else if (opt[1] == "r") vdu = dureal;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "g")       // Fiellerの式使用指定
  {
    if (opt[1].empty() || opt[1] == "+") vg = true;
    else if (opt[1] == "-") vg = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "y")       // 用量を推定する反応値の指定
  {
    double fdata = 0.0;
    if (opt[1].empty() || opt[1] == "m") vy = BLANKf;
    else if (opt[1] == "-") vy = -DBL_MAX;
    else vy = fdata = strtodf(opt[1]);
    if (fdata == BLANKf) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t nd)   // オプションのチェックとセット
{
  if (set_ngp() < 0) return -2; // -resオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  if (clst.empty() || nd < 2) return 0; // 最大対比法実行せず
  if (clst.size() > nd) return disperr(vpname, "Too many coefficients of maximum contrast method !");
  vcvec.resize(nd);
  vcvec.assign(nd, 0.0);        // 初期値を0にする
  size_t i = 0;
  double sum = 0.0;
  for (lstr_citr citr = clst.begin(); citr != clst.end(); ++citr, ++i)
  {
    double fdata = strtodf(*citr);
    if (fdata == BLANKf) return disperr(vpname, "Wrong coefficient of maximum contrast method->" + *citr);
    vcvec[i] = fdata;
    sum += fdata;
  }
  if (sgn(sum) != 0) return disperr(vpname, "Sum of coefficients must be 0 !");
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["drl"] = vngpmap["mse"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["drl"] = vngpmap["mse"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngppos = vngpmap.find(ngp);
      if (ngppos == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = vresmap["y"] = true;
  if (vreslst.size() > 0)
  {
     vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = vresmap["y"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all")
      {
        vresmap["def"] = vresmap["cc"] = vresmap["reg"] = vresmap["stat"] = vresmap["y"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, " Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "ym")
      { vresmap["y"] = true;  vy = BLANKf; }
      else if (res[0] == 'y' && res.size() > 1)
      {
        vy = strtodf(res.substr(1)); // 反応値
        if (vy == BLANKf) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["y"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  int vnd;                      // 用量数
  mapds vdsmap;                 // 用量別統計量マップ
  statcor<double> dstatc;       // 用量と反応の相関基礎統計量
  double vsd;                   // 用量平方和
  double vsb;                   // 回帰平方和
  double vsl;                   // ズレ平方和
  double vsr;                   // 残差平方和
  double va;                    // 用量反応直線の定数
  double vb;                    // 用量反応直線の回帰係数
  double vseb;                  // 回帰係数の標準誤差
  int cal_drline(const option& opt); // 用量-反応直線を計算する
public:
  varclass() : vnd(0), vdsmap(), dstatc(), vsd(0.0), vsb(0.0), vsl(0.0), vsr(0.0), va(0.0), vb(0.0), vseb(0.0) // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int nd() const                // 用量数を返す
  { return vnd; }
  const mapds& dsmap() const    // 用量別統計量マップを返す
  { return vdsmap; }
  int n() const                 // 用量の例数を返す
  { return dstatc.n(); }
  double mx() const             // 用量の平均値を返す
  { return dstatc.mx(); }
  double my() const             // 反応の平均値を返す
  { return dstatc.my(); }
  double sxx() const            // 用量平方和を返す
  { return dstatc.sxx(); }
  double syy() const            // 反応平方和を返す
  { return dstatc.syy(); }
  double sdy() const            // 反応標準偏差を返す
  { return dstatc.sdy(); }
  double sey() const            // 反応標準誤差を返す
  { return dstatc.sey(); }
  double sxy() const            // 用量と反応積和を返す
  { return dstatc.sxy(); }
  double sd() const             // 用量平方和を返す
  { return vsd; }
  double sb() const             // 回帰平方和を返す
  { return vsb; }
  double sl() const             // ズレ平方和を返す
  { return vsl; }
  double sr() const             // 残差回帰平方和を返す
  { return vsr; }
  double a() const              // 用量反応直線の定数を返す
  { return va; }
  double b() const              // 用量反応直線の回帰係数を返す
  { return vb; }
  double seb() const            // 回帰係数の標準誤差を返す
  { return vseb; }
  int cal(const option& opt, ldata datalst);
  // データを読み込み用量反応解析を行う
};

int varclass::cal_drline(const option& opt) // 用量-反応直線を計算する
{
  vsd = 0.0;  vnd = 0;
  for (mapds_citr ditr = vdsmap.begin(); ditr != vdsmap.end(); ++ditr)
  {
    double dose = ditr->first;
    if (opt.du() != dulog || sgn(dose) > 0)
    {
      ++vnd;
      vsd += ditr->second.n() * ditr->second.mean() * ditr->second.mean();
    }
  }
  if (vnd < 2) return vnd;      // 用量-反応直線計算できず
  double eps = DBL_EPSILON * fmax(5.0, vsd);
  vsd -= dstatc.n() * dstatc.my() * dstatc.my();
  if (vsd < eps) vsd = 0.0;
  vsr = dstatc.syy() - vsd;
  if (vsr < eps) vsr = 0.0;
  double sxx = dstatc.sxx(), sxy = dstatc.sxy();
  vsb = (sxx > 0.0) ? sxy * sxy / sxx : 0.0;
  if (vsb < eps) vsb = 0.0;
  vsl = vsd - vsb;
  if (vsl < eps) vsl = 0.0;
  int dft = dstatc.n() - 1, dfd = vnd - 1, dfb = 1;
  int dfr = dft - dfd;
  vb = (sxx > 0.0) ? sxy / sxx : 0.0;
  va = dstatc.my() - vb * dstatc.mx();
  if (dfr < 1)                  // 用量データが1例のみの時
  {
    dfr = dft -  dfb;
    vsr = vsd - vsb;
  }
  vseb = (dfr > 0 && sxx > 0.0) ? sqrt(vsr / (double)dfr / sxx) : 0.0;
  return vnd;                   // 用量数を返す
}

int varclass::cal(const option& opt, ldata datalst)
// データを読み込み用量反応解析を行う
{
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double dose = ditr->fdata(0), res = ditr->fdata(1);
    if (dose != BLANKf && res != BLANKf) // dose=0も集計する
    {
      mapds_itr dsitr = vdsmap.find(dose);
      if (dsitr == vdsmap.end()) // 新規用量
      {
        statv<double> statvi(res);
        vdsmap.insert(make_pair(dose, statvi));
      }
      else                      // すでにある用量
      { dsitr->second.sigma(res);  }
      if (opt.du() != dulog || dose > 0.0)
      {
        double mdose = (opt.du() == dulog) ? log10(dose) : dose;
        dstatc.sigma(mdose, res);
      }
    }
  }
  return cal_drline(opt);       // 用量数を返す
}

class doserclass                // doserメインクラス
{
private:
  string fname;                 // ファイル名
  void put_resid(ostream& fout, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_mstat(ostream& fout, const option& opt, const varclass& var) const;
  // 用量別基礎統計量を出力する
  void put_estd(ostream& fout, option& opt, const varclass& var) const;
  // 用量を逆推定する
  void put_drtab(ostream& fout, const varclass& var) const;
  // 用量反応用分散分析表を出力する
  void put_maxcomp(ostream& fout, const option& opt, const varclass& var) const;
  // 最大対比法結果を出力する
  void plot_ngp(option& opt, idd& id, const varclass& var) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngpdrl(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に用量-反応直線を出力する
  void plot_ngpmse(ofstream& fout, const option& opt, const varclass& var) const;
  // Ngraph用に平均±SEを出力する
public:
  doserclass() : fname()        // コンストラクタ
  {}
  ~doserclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void doserclass::put_resid(ostream& fout, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "DOSE ITEM    :" << id.item(0) << "\n";
  fout << "RESPONSE ITEM:" << id.item(1) << endl;
}

void doserclass::put_mstat(ostream& fout, const option& opt, const varclass& var) const
// 用量別基礎統計量を出力する
{
  const int LENline = (opt.du() == dureal) ? 54 : 70;
  fout << "\n    Dose";
  if (opt.du() == dureal) fout << " ";
  else fout << "     x=log(Dose)";
  fout << "         n        mean          SD          SE\n";
  putline(fout, LENline);
  for (mapds_citr ditr = (var.dsmap()).begin(); ditr != (var.dsmap()).end(); ++ditr)
  {
    fout << setw(8) << ditr->first;
    if (opt.du() == dulog)
    {
      if (sgn(ditr->first) > 0) fout << setw(16) << log10(ditr->first);
      else fout << "     (exclusion)";
    }
    fout << setw(10) << ditr->second.n() << setw(12) << ditr->second.mean();
    fout << setw(12) << ditr->second.sd() << setw(12) << ditr->second.se();
    fout << "\n";
  }
  putline(fout, LENline);
  if (opt.du() == dureal)
  { fout << setw(8) << var.mx(); }
  else
  {
    string tstr("MEAN");
    int npad = tstr.size() - strclen(tstr, CHARwide);
    fout << setw(8 + npad) << tstr << setw(16) << var.mx();
  }
  fout << setw(10) << var.n() << setw(12) << var.my();
  fout << setw(12) << var.sdy() << setw(12) << var.sey();
  fout << endl;
}

void doserclass::put_estd(ostream& fout, option& opt, const varclass& var) const
// 用量を逆推定する
{
  if (var.b() == 0.0) return;   // 用量の逆推定できず
  double y = (opt.y() == BLANKf) ? var.my() : opt.y();
  double cf = (y - var.my()) / var.b();
  double x = var.mx() + cf;
  fout << "Dose at Response(y)=" << y << ":";
  if (opt.du() == dureal) fout << "x=" << x << "\n";
  else fout << "D=" << pow(10.0, x) << "(x=log(D)=" << x << ")\n";
  if (opt.res("cc") == false) return;
  int dft = var.n() - 1, dfd = var.nd() - 1, dfb = 1;
  int dfr = dft - dfd;
  if (dfr < 1) dfr = dft -  dfb;
  double vr = (dfr > 0) ? var.sr() / (double)dfr : 0.0;
  double t = ptot(1.0 - opt.cc(), dfr);
  double g = t / var.b() * t / var.b() * vr / var.sxx();
  if (opt.g() == false || g >= 1.0) g = 0.0; // g>1だと計算できないため
  double gr = 1.0 - g;
  double v = vr * (gr * (1.0 + 1.0 / (double)(var.n())) + cf * cf / var.sxx());
  double ci = t / var.b() * sqrt(v);
  double xl = (x - g * var.mx() - ci) / gr, xu = (x - g * var.mx() + ci) / gr;
  fout << opt.cc() * 100.0 << "%CI of Dose:";
  if (opt.du() == dureal) fout << "Lower=" << xl << "  Upper=" << xu << "\n";
  else fout << "Lower=" << pow(10.0, xl) << "(x=" << xl << ")  Upper=" << pow(10.0, xu) << "(x=" << xu << ")\n";
}

void doserclass::put_drtab(ostream& fout, const varclass& var) const
// 用量反応用分散分析表を出力する
{
  constexpr int LENline = 74;
  string tstr("ANOVA Table");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR            SS        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  int dft = var.n() - 1, dfd = var.nd() - 1, dfb = 1;
  int dfr = dft - dfd;
  if (dfr < 1) dfr = dfd - dfb;
  double v = 0.0, f = 0.0, pr = 0.0;
  calvfp(v, f, pr, var.sb(), dfb, var.sr(), dfr);
  fout << "REGRESSION" << setw(10) << var.sb() << setw(10) << dfb;
  fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  int dfl = dfd - dfb;
  calvfp(v, f, pr, var.sl(), dfl, var.sr(), dfr);
  fout << "LOF" << setw(17) << var.sl() << setw(10) << dfl << setw(14) << v;
  if (dft > dfd) fout << setw(12) << f << setw(14) << pr;
  fout << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sd(), dfd, var.sr(), dfr);
  fout << "DOSE" << setw(16) << var.sd() << setw(10) << dfd;
  if (dft > dfd)
  {
    fout << setw(14) << v << setw(12) << f;
    fout << setw(14) << pr << pind(pr) << "\n";
    v = (dfr > 0) ? var.sr() / (double)dfr : 0.0;
    fout << "RESIDUALS" << setw(11) << var.sr() << setw(10) << dfr;
    fout << setw(14) << v << "\n";
    putline(fout, LENline);
    fout << "TOTAL" << setw(15) << var.syy() << setw(10) << dft;
  }
  fout << endl;
}

void doserclass::put_maxcomp(ostream& fout, const option& opt, const varclass& var) const
// 最大対比法結果を出力する
{
  fout << "\n<MAXIMUM CONTRAST METHOD>\n";
  double sp = 0.0, div = 0.0;
  int nd = 0;
  fout << "Coefficient:";
  for (mapds_citr ditr = (var.dsmap()).begin(); ditr != (var.dsmap()).end(); ++ditr)
  {
    double dose = ditr->first;
    if (opt.du() != dulog || sgn(dose) > 0)
    {
      fout << "dose" << dose << "=" << opt.c(nd) << "  ";
      sp += ditr->second.n() * opt.c(nd) * ditr->second.mean();
      div += ditr->second.n() * opt.c(nd) * opt.c(nd);
      ++nd;
    }
  }
  fout << "\nContribution Rate:r-squared=" << sp / div * sp / var.sd() << "\n";
  int dft = var.n() - 1, dfd = var.nd() - 1;
  int dfr = dft - dfd;
  double vr = (dfr > 0) ? var.sr() / (double)dfr : 0.0;
  double f = (vr > 0.0) ? sp / div * sp / vr : 0.0;
  ldbl pr = ftop(f, 1, dfr);
  fout << "Fisher type :F=" << f << " DF1=1  DF2=" << dfr << "  p=" << pr << pind(pr) <<  "\n";
  f /= (double)dfd;
  pr = ftop(f, dfd, dfr);
  fout << "Scheffe type:F=" << f << " DF1=" << dfd << "  DF2=" << dfr << "  p=" << pr << pind(pr) << endl;
}

void doserclass::plot_ngp(option& opt, idd& id, const varclass& var) const
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  bool delml = false;
  if (opt.ngp("drl"))           // 用量反応直線
  {
    fout << "-s1 -dline -w80";
    if (opt.du() == dulog) fout << " -exlog";
    fout << "\n";
    plot_ngpid(fout, id);
    plot_ngpdrl(fout, opt, var);
    delml = true;
  }
  if (opt.ngp("mse"))           // 平均値±SE
  {
    if (delml)                   // 区切り行を出力する
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dmark,00 -o300";
      if (opt.du() == dulog) fout << " -exlog";
      fout << "\n";
    }
    plot_ngpid(fout, id);
    plot_ngpmse(fout, opt, var);
  }
}

void doserclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  fout << "%DOSE ITEM    :" << id.item(0) << "\n";
  fout << "%RESPONSE ITEM:" << id.item(1) << endl;
}

void doserclass::plot_ngpmse(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に平均±SEを出力する
{
  fout << "%<MEAN +/- SE(draw SE with -derrorbar_y -w40 -o500)>\n";
  for (mapds_citr ditr = (var.dsmap()).begin(); ditr != (var.dsmap()).end(); ++ditr)
  {
    if (opt.du() != dulog || sgn(ditr->first) > 0)
    {
      double se = ditr->second.se();
      fout << ditr->first << " " << ditr->second.mean() << " " << -se << " " << se << endl;
    }
  }
}

void doserclass::plot_ngpdrl(ofstream& fout, const option& opt, const varclass& var) const
// Ngraph用に用量-反応直線を出力する
{
  mapds_citr ditrmin = (var.dsmap()).end(), ditrmax = (var.dsmap()).end();
  for (mapds_citr ditr = (var.dsmap()).begin(); ditr != (var.dsmap()).end(); ++ditr)
  {
    if (opt.du() != dulog || sgn(ditr->first) > 0)
    {
      if (ditrmin == (var.dsmap()).end()) ditrmin = ditr;
      ditrmax = ditr;
    }
  }
  fout << "%<DOSE-RESPONSE LINE>\n";
  double xmin = (opt.du() == dureal) ? ditrmin->first : log10(ditrmin->first);
  double xmax = (opt.du() == dureal) ? ditrmax->first : log10(ditrmax->first);
  double range = xmax - xmin;
  double xstart = xmin - range * 0.1, xend = xmax + range * 0.1;
  double y = var.a() + var.b() * xstart;
  double dose = (opt.du() == dulog) ? pow(10.0, xstart) : xstart;
  fout << dose << " " << y << "\n";
  y = var.a() + var.b() * xend;
  dose = (opt.du() == dulog) ? pow(10.0, xend) : xend;
  fout << dose << " " << y << endl;
}

int doserclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (id.ni() != 2) throw runtime_error("Number of item must be 2 !");
  return nc;
}

void doserclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (id.nc() < 1)              // データなし
  {
    if (opt.res("res"))
    {
      put_resid(cout, id);
      cout << "\n" << endl;
    }
    return;
  }
  varclass var;                 // 変数オブジェクト
  int nd = var.cal(opt, datalst);
  if (opt.check(nd) < 0) return; // 最大対比法の係数不適切
  if (opt.res("res"))
  {
    put_resid(cout, id);
    if (opt.res("stat")) put_mstat(cout, opt, var);
    if (opt.res("reg"))
    {
      double r2 = var.sb() / var.sd();
      cout << "\n";
      if (opt.du() == dulog) cout << "log(Dose)";
      else cout << "Dose";
      cout << "-Response Line:y=" << var.a() << showpos << var.b() << noshowpos << "x\n";
      cout << "Contribution Rate:r-squared=" << r2 << "  SE of Regression Coef=" << var.seb() << "\n";
      if (opt.res("y")) put_estd(cout, opt, var);
      if (var.syy() > 0.0) put_drtab(cout, var);
    }
    if (nd > 2 && opt.csize() > 0) put_maxcomp(cout, opt, var);
    cout << "\n" << endl;
  }
  if (opt.ngpname() != "") plot_ngp(opt, id, var);
}

int main(int argc, char *argv[]) // メインルーチン
{
  doserclass doser;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) doser.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { doser.get(argv[i], opt, id, datalst);  fok = false; }
    }
    doser.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
