// parale.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 17:00:29 tac>
//
// 最初の項目を群項目、次の項目を用量項目、最後の項目を反応項目とした
// 平行線検定を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W parale parale.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "parale  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:paral [<OPTION> ...]  [<DATA FILE> ...]",
  "  The parallel line assay with the first item as group and second item as",
  "  dose and third item as response.",
  "<OPTION>",
  "  -du={l|r}:select the unit of dose",
  "    l:log(dose)(default)  r:dose",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -res=[-][{def|all|at|cc[<C-COEF>]|er|g|ma[{b|f|s|t}]|stat}][,...]",
  "    :select output results",
  "    -:wiht -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    at:ANOVA table  cc[<C-COEF>]:confidence interval(0=<C-COEF<1 cc:cc0.95)",
  "    er:potency ratio  g:specify whether to use the Fieller's formula",
  "    ma[{b|f|s|t}]:multiple comparison of adjusted means",
  "      b:Bonferroni type  f:Fisher tyep  s:Scheffe type  t:Tukey type",
  "    stat:basic statistics",
  "    default:automatic judgement according to data"
};

const string Title = "=== Parallel Line Assay ==="; // 表題

using mapds = map<double, statv<double>>; // 用量別統計量マップの別名定義
using mapds_itr = mapds::iterator;        // 反復子の別名定義
using mapds_citr = mapds::const_iterator; // const反復子の別名定義

enum dutype                     // 用量単位の種類
{
  dupnon,                       // 単位なし
  dulog,                        // 対数
  dureal,                       // 実数
  duetc                         // その他
};

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  dutype vdu;                   // 用量の単位
  mctype vma;                   // 修正平均値の多重比較法
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  dutype du() const             // 用量の単位を返す
  { return vdu; }
  mctype ma() const             // 修正群差の多重比較法を返す
  { return vma; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェック
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vdu(dulog), vma(mcbon), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "du")           // 用量の単位指定
  {
    if (opt[1].empty() || opt[1] == "l") vdu = dulog;
    else if (opt[1] == "r") vdu = dureal;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェック
{
  if (ni != 3) return disperr(vpname, "項目数は3つだでよ！");
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["er"] = vresmap["g"] = vresmap["stat"] = true;
  vresmap["ma"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["er"] = vresmap["g"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["er"] = vresmap["g"] = vresmap["stat"] = true;
      }
      else if (res == "all")
      {
        vresmap["res"] = vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["er"] = vresmap["g"] = vresmap["ma"] = vresmap["stat"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "mab")
      { vresmap["ma"] = true;  vma = mcbon; }
      else if (res == "maf")
      { vresmap["ma"] = true;  vma = mcfis; }
      else if (res == "mas")
      { vresmap["ma"] = true;  vma = mcsch; }
      else if (res == "mat")
      { vresmap["ma"] = true;  vma = mctuk; }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class gstat                     // 群別統計量クラス
{
private:
  mapds dsmap;                  // 用量別統計量マップ
  vector<mapds_itr> dsitrvec;   // 用量別統計量反復子ベクトル
  statcor<double> dstatc;       // 相関用基礎統計量
  int vnd;                      // 用量数
  double vsa;                   // 用量平方和
  double vsb;                   // 回帰平方和
  double va;                    // 群別定数
  double vb;                    // 群別回帰係数
  double vac;                   // 共通定数
  double vbc;                   // 共通回帰係数
public:
  gstat() : dsmap(), dsitrvec(), dstatc(), vnd(0), vsa(0.0), vsb(0.0), va(0.0), vb(0.0), vac(0.0), vbc(0.0)
  // デフォールトコンストラクタ
  {}
  gstat(dutype du, double dose, double r); // コンストラクタ
  gstat(const gstat& gstat0) : dsmap(gstat0.dsmap), dsitrvec(gstat0.dsitrvec), dstatc(gstat0.dstatc), vnd(gstat0.vnd), vsa(gstat0.vsa), vsb(gstat0.vsb), va(gstat0.va), vb(gstat0.vb), vac(gstat0.vac), vbc(gstat0.vbc)
  // コピーコンストラクタ
  {}
  virtual ~gstat()              // デストラクタ
  {}
  const gstat& operator=(const gstat& gstat0); // コピー代入演算子
  const gstat& operator+=(const gstat& gstat0); // 加算代入演算子
  int n() const                                 // 相関用例数を返す
  { return dstatc.n(); }
  double mx() const             // 用量平均を返す
  { return dstatc.mx(); }
  double my() const             // 反応平均を返す
  { return dstatc.my(); }
  double sxx() const            // 用量積和を返す
  { return dstatc.sxx(); }
  double syy() const            // 反応積和を返す
  { return dstatc.syy(); }
  double sxy() const            // 用量と反応の積和を返す
  { return dstatc.sxy(); }
  double sdy() const            // 反応の標準偏差を返す
  { return dstatc.sdy(); }
  double sey() const            // 反応の標準誤差を返す
  { return dstatc.sey(); }
  double rxy() const            // 用量と反応の相関係数を返す
  { return dstatc.rxy(); }
  int nd() const                // 用量数を返す
  { return vnd; }
  double sa() const             // 用量平方和を返す
  { return vsa; }
  double sb() const             // 回帰平方和を返す
  { return vsb; }
  double a() const              // 群別定数を返す
  { return va; }
  double b() const              // 群別回帰係数を返す
  { return vb; }
  double ac() const             // 共通定数を返す
  { return vac; }
  double ac(double ac0)         // 共通定数をセットする
  { return (vac = ac0); }
  double bc() const             // 共通回帰係数を返す
  { return vbc; }
  double bc(double bc0)         // 共通回帰係数をセットする
  { return (vbc = bc0); }
  double dose(int idx) const    // 用量を返す
  { return dsitrvec[idx]->first; }
  int n(int idx) const          // 反応例数を返す
  { return dsitrvec[idx]->second.n(); }
  double mean(int idx) const    // 反応平均値を返す
  { return dsitrvec[idx]->second.mean(); }
  double vr(int idx) const      // 反応分散を返す
  { return dsitrvec[idx]->second.vr(); }
  double sd(int idx) const      // 反応標準偏差を返す
  { return dsitrvec[idx]->second.sd(); }
  double se(int idx) const      // 応標準誤差を返す
  { return dsitrvec[idx]->second.se(); }
  void sigma(dutype du, double dose, double r); // 用量と反応データを累積する
  int set();                    // 用量別統計量を整理する
  void cal();                   // 群別回帰直線を計算する
};

gstat::gstat(dutype du, double dose, double r) : dsmap(), dsitrvec(), dstatc(), vnd(1), vsa(0.0), vsb(0.0), va(0.0), vb(0.0), vac(0.0), vbc(0.0)
// コンストラクタ
{
  double x = (du == dulog) ? log10(dose) : dose;
  dstatc.sigma(x, r);
  statv<double> statvi(r);
  dsmap.insert(make_pair(dose, statvi));
}

const gstat& gstat::operator=(const gstat& gstat0) // コピー代入演算子
{
  if (&gstat0 == this) return *this; // 自分自身への代入を防ぐ
  dsmap = gstat0.dsmap;  dsitrvec = gstat0.dsitrvec;  dstatc = gstat0.dstatc;
  vnd = gstat0.vnd;  vsa = gstat0.vsa;  vsb = gstat0.vsb;
  va = gstat0.va;  vb = gstat0.vb;
  vac = gstat0.vac;  vbc = gstat0.vbc;
  return *this;                 // 自分自身への参照を返す
}

const gstat& gstat::operator+=(const gstat& gstat0) // 加算代入演算子
{
  if (gstat0.nd() < 1) return *this; // 加算しない
  dstatc += gstat0.dstatc;  vnd += gstat0.vnd;
  vsa += gstat0.vsa;  vsb += gstat0.vsb;
  return *this;                 // 自分自身への参照を返す
}

void gstat::sigma(dutype du, double dose, double r) // 用量と反応データを累積する
{
  mapds_itr ditr = dsmap.find(dose);
  if (ditr == dsmap.end())      // 新規用量
  {
    statv<double> statvi(r);
    dsmap.insert(make_pair(dose, statvi));
  }
  else                      // すでにある用量
  { ditr->second.sigma(r);  }
  double x = (du == dulog) ? log10(dose) : dose;
  dstatc.sigma(x, r);
}

int gstat::set()                // 用量別統計量を整理する
{
  vnd = 0;  vsa = 0.0;
  if (dsmap.empty()) return 0;  // 用量別統計量なし
  dsitrvec.resize(dsmap.size());
  for (mapds_itr ditr = dsmap.begin(); ditr != dsmap.end(); ++ditr, ++vnd)
  {
    dsitrvec[vnd] = ditr;
    vsa += ditr->second.n() * ditr->second.mean() * ditr->second.mean();
  }
  return vnd;                   // 用量数を返す
}

void gstat::cal()               // 群別回帰直線を計算する
{
  vsa -= dstatc.n() * dstatc.my() * dstatc.my();
  const double Veps = DBL_EPSILON * fmax(5.0, dstatc.syy());
  if (vsa < Veps) vsa = 0.0;
  double sxx = dstatc.sxx(), sxy = dstatc.sxy();
  vsb = (sxx > 0.0) ? sxy * sxy / sxx : 0.0;
  vb = (sxx > 0.0) ? sxy / sxx : 0.0;
  va = dstatc.my() - vb * dstatc.mx();
}

using mapsg = map<string, gstat>;         // 群別統計量マップの別名定義
using mapsg_itr = mapsg::iterator;        // 反復子の別名定義
using mapsg_citr = mapsg::const_iterator; // const反復子の別名定義

class varclass                  // 変数クラス
{
private:
  int vng;                      // 群数
  mapsg vgsmap;                 // 群別統計量マップ
  gstat gstatall;               // 全体の統計量用
  double vsxxc;                 // 全体の共通用量平方和
  double vsp;                   // 全体の群差平方和
  double vsbc;                  // 全体の共通回帰平方和
  double vsid;                  // 全体の修正群差平方和
  double vsb0;                  // 全体の回帰平方和
  double vsdp;                  // 全体の非平行性平方和
  double vsr;                   // 全体の残差平方和
  int vdfr;                     // 全体の残差自由度
  void cal_paral();             // 平行線検定用計算
public:
 varclass() : vng(0), vgsmap(), gstatall(), vsxxc(0.0), vsp(0.0), vsbc(0.0), vsid(0.0), vsb0(0.0), vsdp(0.0), vsr(0.0), vdfr(0) // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int ng() const                // 群数を返す
  { return vng; }
  const mapsg& gsmap() const    // 群別統計量マップを返す
  { return vgsmap; }
  int n() const                 // 全体の例数を返す
  { return gstatall.n(); }
  int nd() const                // 全体の用量数を返す
  { return gstatall.nd(); }
  double mx() const             // 全体の用量平均を返す
  { return gstatall.mx(); }
  double my() const             // 全体の反応平均を返す
  { return gstatall.my(); }
  double sxx() const            // 全体の用量積和を返す
  { return gstatall.sxx(); }
  double syy() const            // 全体の総平方和を返す
  { return gstatall.syy(); }
  double sxy() const            // 全体の用量と反応の積和を返す
  { return gstatall.sxy(); }
  double a() const              // 全体の反応直線の定数を返す
  { return gstatall.a(); }
  double b() const              // 全体の反応直線の回帰係数を返す
  { return gstatall.b(); }
  double ac() const             // 全体の共通反応直線の定数を返す
  { return gstatall.ac(); }
  double bc() const             // 全体の共通反応直線の回帰位数を返す
  { return gstatall.bc(); }
  double sa() const             // 全体の用量平方和を返す
  { return gstatall.sa(); }
  double sb() const             // 全体の回帰平方和を返す
  { return gstatall.sb(); }
  double sxxc() const           // 全体の共通用量平方和を返す
  { return vsxxc; }
  double sp() const             // 全体の群差平方和を返す
  { return vsp; }
  double sbc() const            // 全体の共通回帰平方和を返す
  { return vsbc; }
  double sid() const            // 全体の修正群差平方和を返す
  { return vsid; }
  double sb0() const            // 全体の回帰平方和を返す
  { return vsb0; }
  double sdp() const            // 全体の非平行性平方和を返す
  { return vsdp; }
  double sr() const             // 全体の残差平方和を返す
  { return vsr; }
  int dfr() const               // 全体の残差自由度を返す
  { return vdfr; }
  int cal(const option& opt, ldata datalst);
  // データを読み込み平行線検定を行う
};

void varclass::cal_paral()      // 平行線検定用計算
{
  vsxxc = 0.0;  vsbc = 0.0;  vsb0 = 0.0;  vsid = 0.0;  vsp = 0.0;  vsdp = 0.0;
  vsr = 0.0;  vdfr = 0;
  double syyc = 0.0, sxyc = 0.0;
  int dfd = -1;
  for (mapsg_citr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr)
  {
    vsxxc += gitr->second.sxx();  syyc += gitr->second.syy();
    sxyc += gitr->second.sxy();
    vsp += gitr->second.n() * gitr->second.my() * gitr->second.my();
    vsdp += gitr->second.b() * gitr->second.sxy();
    dfd += gitr->second.nd();
  }
  gstatall.bc((vsxxc > 0.0) ? sxyc / vsxxc : 0.0);
  gstatall.ac(gstatall.my() - gstatall.bc() * gstatall.mx());
  vsp -= gstatall.n() * gstatall.my() * gstatall.my();
  const double Seps = DBL_EPSILON * fmax(5.0, syyc);
  if (vsp < Seps) vsp = 0.0;
  vsbc = sxyc * gstatall.bc();
  if (vsbc < Seps) vsbc = 0.0;
  vsb0 = gstatall.b() * gstatall.sxy();
  if (vsb0 < Seps) vsb0 = 0.0;
  vsid = vsp + vsbc - vsb0;
  vsdp -= vsbc;
  if (vsid < Seps) vsid = 0.0;
  if (vsdp < Seps) vsdp = 0.0;
  vdfr = gstatall.n() - 1 - dfd;
  vsr = gstatall.syy() - gstatall.sa();
  if (vsr < Seps) vsr = 0.0;
  for (mapsg_itr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr)
  {
    gitr->second.bc(gstatall.bc());
    gitr->second.ac(gitr->second.my() - gitr->second.bc() * gitr->second.mx());
  }
}

int varclass::cal(const option& opt, ldata datalst)
// データを読み込み平行線検定を行う
{
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double dose = ditr->fdata(1), r = ditr->fdata(2);
    if (ditr->sdata(0) != BLANKs && dose != BLANKf && (opt.du() != dulog || sgn(dose) > 0) && r != BLANKf)
    {
      mapsg_itr itr = vgsmap.find(ditr->sdata(0));
      if (itr == vgsmap.end())  // 新規群
      {
        gstat gstati(opt.du(), dose, r);
        vgsmap.insert(make_pair(ditr->sdata(0), gstati));
      }
      else                      // すでにある群
      { itr->second.sigma(opt.du(), dose, r);  }
    }
  }
  for (mapsg_itr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr)
  {
    gitr->second.set();
    gstatall += gitr->second;
    gitr->second.cal();
  }
  gstatall.cal();
  cal_paral();
  vng = vgsmap.size();
  return vng;                   // 群数を返す
}

class paralclass                // paralメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_dstat(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 群ごとの基礎統計量と(対数)用量-反応直線を出力する
  void put_mstat(ostream& fout, const option& opt, const gstat& gstati) const;
  // 群ごとの用量別基礎統計量を出力する
  void put_paraltab(ostream& fout, const varclass& var) const;
  // 平行線検定用分散分析表を出力する
  void put_erest(ostream& fout, option& opt, const varclass& var) const;
  // 効力比(用量差)推定値を出力する
  void put_macmp(ostream& fout, const option& opt, const varclass& var) const;
  // 修正群差の多重比較を出力する
  void put_maci(ostream& fout, const option& opt, const varclass& var) const;
  // 修正群差の信頼区間を出力する
public:
  paralclass() : fname()        // コンストラクタ
  {}
  ~paralclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void paralclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM   :" << id.item(0) << "\n";
  fout << "DOSE ITEM    :" << id.item(1) << "\n";
  fout << "RESPONSE ITEM:" << id.item(2) << "\n";
}

void paralclass::put_dstat(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 群ごとの基礎統計量と(対数)用量-反応直線を出力する
{
  string str("Dose");
  if (opt.du() == dulog) str = "log(Dose)";
  double sb = 0.0, sl = 0.0, v = 0.0, f = 0.0, pr = 0.0;
  int dfb = 1, dfl = 0, i = 1;
  const double Seps = DBL_EPSILON * fmax(5.0, var.syy());
  for (mapsg_citr gitr = (var.gsmap()).begin(); gitr != (var.gsmap()).end(); ++gitr, ++i)
  {
    fout << "\nGROUP" << setw(2) << i;
    fout << ":" << id.item(0) << "=" << gitr->first << "\n";
    put_mstat(fout, opt, gitr->second);
    fout << "Group " << str << "-Response Line:y=" << gitr->second.a();
    fout << showpos << gitr->second.b() << noshowpos << "x\n";
    fout << "Common " << str << "-Response Line:y=" << gitr->second.ac();
    fout << showpos << gitr->second.bc() << noshowpos << "x\n";
    double r2 = (gitr->second.sa() > 0.0) ? gitr->second.sb() / gitr->second.sa() : 0.0;
    fout << "Contribution Rate of Group " << str << "-Response Line:r-squared=" << r2 << "\n";
    sb = gitr->second.sb();
    calvfp(v, f, pr, sb, dfb, var.sr(), var.dfr());
    fout << "Test of Group " << str << "-Response Line:p=" << pr << pind(pr) << "\n";
    dfl = gitr->second.nd() - 1 - dfb;
    sl = gitr->second.sa() - sb;
    if (dfl < 1 || sl < Seps) sl = 0.0;
    calvfp(v, f, pr, sl, dfl, var.sr(), var.dfr());
    fout << "Test of LOF(Lack Of Fit):p=" << pr << pind(pr) << "\n";
  }
  if (var.ng() < 2) return;
  fout << "\n<OVERALL>\n";
  fout << "Overall " << str << "-Response Line:y=" << var.a();
  fout << showpos << var.b() << noshowpos << "x\n";
  fout << "Common " << str << "-Response Line:y=" << var.ac();
  fout << showpos << var.bc() << noshowpos << "x\n";
  double r2 = (var.sa() > 0.0) ? var.sb() / var.sa() : 0.0;
  fout << "Contribution Rate of Overall " << str << "-Response line:r-squared=" << r2 << endl;
}

void paralclass::put_mstat(ostream& fout, const option& opt, const gstat& gstati) const
// 群ごとの用量別基礎統計量を出力する
{
  const int LENline = (opt.du() == dureal) ? 54 : 70;
  fout << "    Dose";
  if (opt.du() == dulog) fout << "     x=log(Dose)";
  fout << "         n        mean          SD          SE\n";
  putline(fout, LENline);
  for (int j = 0; j < gstati.nd(); ++j)
  {
    fout << setw(8) << gstati.dose(j);
    if (opt.du() == dulog) fout << setw(16) << log10(gstati.dose(j));
    fout << setw(10) << gstati.n(j) << setw(12) << gstati.mean(j);
    fout << setw(12) << gstati.sd(j) << setw(12) << gstati.se(j) << "\n";
  }
  putline(fout, LENline);
  if (opt.du() == dureal)
  { fout << setw(8) << gstati.mx(); }
  else
  {
    string tstr("MEAN");
    int npad = tstr.size() - strclen(tstr, CHARwide);
    fout << setw(8 + npad) << tstr << setw(16) << gstati.mx();
  }
  fout << setw(10) << gstati.n() << setw(12) << gstati.my();
  fout << setw(12) << gstati.sdy() << setw(12) << gstati.sey() << endl;
}

void paralclass::put_paraltab(ostream& fout, const varclass& var) const
// 平行線検定用分散分析表を出力する
{
  if (var.syy() <= 0.0) return; // 出力なし
  constexpr int LENline = 78;
  int dfp = var.ng() - 1, dfd = -1;
  int dft = var.n() - 1, dfb = 1;
  for (mapsg_citr gitr = (var.gsmap()).begin(); gitr != (var.gsmap()).end(); ++gitr)
  { dfd += gitr->second.nd(); }
  string tstr("ANOVA Table");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, var.sp(), dfp, var.sr(), var.dfr());
  fout << "GROUP" << setw(19) << var.sp() << setw(10) << dfp << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  double sla = 0.0;
  int dfla = 0, i = 1;
  const double Seps = DBL_EPSILON * fmax(5.0, var.syy());
  for (mapsg_citr gitr = (var.gsmap()).begin(); gitr != (var.gsmap()).end(); ++gitr, ++i)
  {
    double sb = gitr->second.sb();
    int dfl = gitr->second.nd() - 1 - dfb;
    double sl = gitr->second.sa() - sb;
    if (dfl < 1 || sl < Seps) sl = 0.0;
    sla += sl;  dfla += dfl;
  }
  calvfp(v, f, pr, var.sbc(), dfb, var.sr(), var.dfr());
  fout << "COMMON-REG" << setw(14) << var.sbc() << setw(10) << dfb << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sid(), dfp, var.sr(), var.dfr());
  fout << "ADJ-GROUP" << setw(15) << var.sid() << setw(10) << dfp << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, var.sb0(), dfb, var.sr(), var.dfr());
  fout << "OVERALL-REG" << setw(13) << var.sb0() << setw(10) << dfb << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, sla, dfla, var.sr(), var.dfr());
  fout << "TOTAL LOF" << setw(15) << sla << setw(10) << dfla << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, var.sdp(), dfp, var.sr(), var.dfr());
  fout << "NONPARALLEL" << setw(13) << var.sdp() << setw(10) << dfp << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sa(), dfd, var.sr(), var.dfr());
  fout << "DOSE" << setw(20) << var.sa() << setw(10) << dfd;
  fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  fout << "RESIDUALS" << setw(15) << var.sr() << setw(10) << var.dfr() << setw(14) << vr << "\n";
  putline(fout, LENline);
  fout << "TOTAL" << setw(19) << var.syy() << setw(10) << dft << endl;
}

void paralclass::put_erest(ostream& fout, option& opt, const varclass& var) const
// 効力比(用量差)推定値を出力する
{
  double bc = var.bc();
  if (sgn(bc) == 0) return;     // 効力比推定できず
  int ng = var.ng();
  int dc = ng * (ng - 1) / 2;
  double vr = (var.dfr() > 0) ? var.sr() / var.dfr() : 0.0;
  if (opt.du() == dureal) fout << "\n<Dose Differences d";
  else fout << "\n<Potency Ratio R";
  fout << " and " << opt.cc() * 100.0 << "%";
  if (dc > 1) fout << " Bonferroni type Simultaneous";
  fout << " Confidence Interval";
  if (opt.res("g") == false) fout << "(Approximate Calculation)";
  fout << ">\n";
  double t = ptot((1.0 - opt.cc()) / (double)dc, var.dfr());
  double bcx = bc * bc * var.sxxc();
  double g = (opt.res("g")) ? t * t * vr / bcx : 0.0;
  if (g >= 1.0) g = 0.0;        // g>1だと計算できないため
  double gr = 1.0 - g;
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 1; i < ng; ++i, ++gitri)
  {
    double wi = 1.0 / (double)(gitri->second.n());
    double mxi = gitri->second.mx();
    double myi = gitri->second.my(), aci = gitri->second.ac();
    mapsg_citr gitrj = gitri;
    ++gitrj;
    int j = i + 1;
    while (gitrj != (var.gsmap()).end())
    {
      double w = wi + 1.0 / (double)(gitrj->second.n());
      double dm = gitrj->second.ac() - aci;
      double mdx = mxi - gitrj->second.mx();
      double mdy = myi - gitrj->second.my();
      double m = dm / bc;
      if (opt.du() == dureal) fout << "GROUP" << i << " - GROUP" << j << ":d=" << m;
      else fout << "GROUP" << j << "/GROUP" << i << ":R=" << pow(10.0, m) << " (log(R)=" << m << ")\n";
      double vm = vr * (gr * w + mdy * mdy / bcx);
      double ci = t / bc * sqrt(vm);
      double ml = (m - g * mdx - ci) / gr, mu = (m - g * mdx + ci) / gr;
      fout << "CI:Lower=";
      if (opt.du() == dureal) fout << ml << "  Upper=" << mu << "\n";
      else fout << pow(10.0, ml) << "(log(R)=" << ml << ")  Upper=" << pow(10.0, mu) << "(log(R)=" << mu << ")\n";
      ++gitrj;  ++j;
    }
  }
}

void paralclass::put_macmp(ostream& fout, const option& opt, const varclass& var) const
// 修正群差の多重比較を出力する
{
  if (var.sr() <= 0.0) return;  // 出力なし
  int ng = var.ng(), df1 = 1, dc = 1;
  string method = "Tukey";
  switch (opt.ma())
  {
  case mcbon : method = "Bonferroni";  dc = ng * (ng - 1) / 2;
    break;
  case mcfis : method = "Fisher";
    break;
  case mcsch : method = "Scheffe";  df1 = ng - 1;
  }
  fout << "\n<" << method << " Type Multiple Comparison of Adjusted Means>\n";
  if (opt.ma() == mctuk)
  { fout << "GROUP   - GROUP        q-value      GROUPS          DF       p-value\n"; }
  else
  { fout << "GROUP   - GROUP        F-value         DF1         DF2       p-value\n"; }
  constexpr int LENline = 72;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = var.sr() / var.dfr();
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < ng - 1; ++i, ++gitri)
  {
    double wi = 1.0 / (double)(gitri->second.n());
    double mxi = gitri->second.mx(), aci = gitri->second.ac();
    mapsg_citr gitrj = gitri;
    int npadi = gitri->first.size() - strclen(gitri->first, CHARwide);
    ++gitrj;
    while (gitrj != (var.gsmap()).end())
    {
      double w = wi + 1.0 / (double)(gitrj->second.n());
      double dm = gitrj->second.ac() - aci, dmx = gitrj->second.mx() - mxi;
      double vd = (w + dmx * dmx / var.sxxc()) * vr;
      double f = dm * dm / vd / (double)df1;
      double pr = calmcpr(opt.ma(), f, ng, df1, var.dfr(), dc);
      int npadj = gitrj->first.size() - strclen(gitrj->first, CHARwide);
      fout << left << setw(8 + npadj) << gitrj->first;
      fout << "- ";
      fout << setw(8 + npadi) << gitri->first << right << setw(12) << f;
      if (opt.ma() == mctuk) fout << setw(12) << ng;
      else fout << setw(12) << df1;
      fout << setw(12) << var.dfr() << setw(14) << pr << pind(pr) << "\n";
      ++gitrj;
    }
  }
  putline(fout, LENline);
}

void paralclass::put_maci(ostream& fout, const option& opt, const varclass& var) const
// 修正群差の信頼区間を出力する
{
  if (var.sr() <= 0.0) return;  // 出力なし
  int ng = var.ng(), df1 = 1, dc = 1;
  string method = "Tukey";
  switch (opt.ma())
  {
  case mcbon : method = "Bonferroni";  dc = ng * (ng - 1) / 2;
    break;
  case mcfis : method = "Fisher";
    break;
  case mcsch : method = "Scheffe";  df1 = ng - 1;
  }
  fout << "\n<" << method << " Type " << opt.cc() * 100.0 << "% Simultaneous Confidence Interval>\n";
  fout << "GROUP   - GROUP   Diff of Mean   Precicion   Lower-lim   Upper-lim\n";
  constexpr int LENline = 66;   // 区切り線の長さ
  putline(fout, LENline);
  double vr = var.sr() / var.dfr();
  mapsg_citr gitri = (var.gsmap()).begin();
  for (int i = 0; i < ng - 1; ++i, ++gitri)
  {
    double wi = 1.0 / (double)(gitri->second.n());
    double mxi = gitri->second.mx(), aci = gitri->second.ac();
    mapsg_citr gitrj = gitri;
    int npadi = gitri->first.size() - strclen(gitri->first, CHARwide);
    ++gitrj;
    while (gitrj != (var.gsmap()).end())
    {
      double w = wi + 1.0 / (double)(gitrj->second.n());
      double dm = gitrj->second.ac() - aci, dmx = gitrj->second.mx() - mxi;
      double vd = (w + dmx * dmx / var.sxxc()) * vr;
      double ci = calmcci(opt.ma(), opt.cc(), ng, df1, var.dfr(), dc, vd);
      int npadj = gitrj->first.size() - strclen(gitrj->first, CHARwide);
      fout << left << setw(8 + npadj) << gitrj->first;
      fout << "- ";
      fout << setw(8 + npadi) << gitri->first << right << setw(12) << dm;
      fout << setw(12) << ci << setw(12) << dm - ci << setw(12) << dm + ci;
      fout << "\n";
      ++gitrj;
    }
  }
  putline(fout, LENline);
}

int paralclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void paralclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (opt.res("res") == false) return;
  put_id(cout, id);
  if (id.nc() < 1)              // データなし
  {
    cout << "\n" << endl;
    return;
  }
  varclass var;                 // 変数オブジェクト
  int ng = var.cal(opt, datalst);
  if (opt.res("stat")) put_dstat(cout, opt, id, var);
  if (opt.res("at")) put_paraltab(cout, var);
  if (ng > 1)
  {
    if (opt.res("er")) put_erest(cout, opt, var);
    if (opt.res("ma"))
    {
      put_macmp(cout, opt, var);
      put_maci(cout, opt, var);
    }
  }
  cout << "\n" << endl;
}

int main(int argc, char *argv[]) // メインルーチン
{
  paralclass paral;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) paral.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { paral.get(argv[i], opt, id, datalst);  fok = false; }
    }
    paral.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
