// slopee.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-08 17:22:48 tac>
//
// 最初の項目を群項目、次の項目を用量項目、最後の項目を反応項目とした
// 勾配比検定を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W slopee slopee.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H,mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "slopee  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:slope [<OPTION> ...]  [<DATA FILE> ...]",
  "  The slope ratio assay with the fist item as group and second item as dose",
  "  and third item response.",
  "<OPTION>",
  "  -f=<FILE>:reads options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -res=[-][{def|all|at|cc[<C-COEF>]|er|stat}][,...]",
  "    :select output results",
  "    -:wiht -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    at:ANOVA table  cc[<C-COEF>]:confidence interval(0=<C-COEF<1 cc:cc0.95)",
  "    er:potency ratio  stat:basic statistics"
};

const string Title = "=== Slope Ratio Assay ==="; // 表題

using mapds = map<double, statv<double>>; // 用量別統計量マップの別名定義
using mapds_itr = mapds::iterator;        // 反復子の別名定義
using mapds_citr = mapds::const_iterator; // const反復子の別名定義

class option                    // オプションクラス
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  bool res(const string& code); // 分析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェック
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェック
{
  if (ni != 3) return disperr(vpname, "項目数は3つだよ！");
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["er"] = vresmap["stat"] = true;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["er"] = vresmap["stat"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def" || res == "all") // 省略時＝全て出力
      {
        vresmap["def"] = vresmap["at"] = vresmap["cc"] = vresmap["er"] = vresmap["stat"] = true;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class gstat                     // 群別統計量クラス
{
private:
  mapds dsmap;                  // 用量別統計量マップ
  vector<mapds_itr> dsitrvec;   // 用量別統計量反復子ベクトル
  statcor<double> dstatc;       // 相関用基礎統計量
  int vnd;                      // 用量数
  double vsa;                   // 用量平方和
  double vsb;                   // 回帰平方和
  double va;                    // 群別定数
  double vb;                    // 群別回帰係数
  double vac;                   // 共通定数
  double vbc;                   // 共通回帰係数
public:
  gstat() : dsmap(), dsitrvec(), dstatc(), vnd(0), vsa(0.0), vsb(0.0), va(0.0), vb(0.0), vac(0.0), vbc(0.0)
  // デフォールトコンストラクタ
  {}
  gstat(double dose, double r); // コンストラクタ
  gstat(const gstat& gstat0) : dsmap(gstat0.dsmap), dsitrvec(gstat0.dsitrvec), dstatc(gstat0.dstatc), vnd(gstat0.vnd), vsa(gstat0.vsa), vsb(gstat0.vsb), va(gstat0.va), vb(gstat0.vb), vac(gstat0.vac), vbc(gstat0.vbc)
  // コピーコンストラクタ
  {}
  virtual ~gstat()              // デストラクタ
  {}
  const gstat& operator=(const gstat& gstat0); // コピー代入演算子
  const gstat& operator+=(const gstat& gstat0); // 加算代入演算子
  int n() const                                 // 相関用例数を返す
  { return dstatc.n(); }
  double mx() const             // 用量平均を返す
  { return dstatc.mx(); }
  double my() const             // 反応平均を返す
  { return dstatc.my(); }
  double sxx() const            // 用量積和を返す
  { return dstatc.sxx(); }
  double syy() const            // 反応積和を返す
  { return dstatc.syy(); }
  double sxy() const            // 用量と反応の積和を返す
  { return dstatc.sxy(); }
  double sdy() const            // 反応の標準偏差を返す
  { return dstatc.sdy(); }
  double sey() const            // 反応の標準誤差を返す
  { return dstatc.sey(); }
  double rxy() const            // 用量と反応の相関係数を返す
  { return dstatc.rxy(); }
  int nd() const                // 用量数を返す
  { return vnd; }
  double sa() const             // 用量平方和を返す
  { return vsa; }
  double sb() const             // 回帰平方和を返す
  { return vsb; }
  double a() const              // 群別定数を返す
  { return va; }
  double b() const              // 群別回帰係数を返す
  { return vb; }
  double ac() const             // 共通定数を返す
  { return vac; }
  double ac(double ac0)         // 共通定数を設定する
  { return (vac = ac0); }
  double bc() const             // 共通回帰係数を返す
  { return vbc; }
  double bc(double bc0)         // 共通回帰係数を設定する
  { return (vbc = bc0); }
  double dose(int idx) const    // 用量を返す
  { return dsitrvec[idx]->first; }
  int n(int idx) const          // 反応例数を返す
  { return dsitrvec[idx]->second.n(); }
  double mean(int idx) const    // 反応平均値を返す
  { return dsitrvec[idx]->second.mean(); }
  double vr(int idx) const      // 反応分散を返す
  { return dsitrvec[idx]->second.vr(); }
  double sd(int idx) const      // 反応標準偏差を返す
  { return dsitrvec[idx]->second.sd(); }
  double se(int idx) const      // 反応標準誤差を返す
  { return dsitrvec[idx]->second.se(); }
  void sigma(double dose, double r); // 用量と反応データを累積する
  int set();                         // 用量別統計量を整理する
  void cal();                        // 群別回帰直線を計算する
};

gstat::gstat(double dose, double r) : dsmap(), dsitrvec(), dstatc(dose, r), vnd(1), vsa(0.0), vsb(0.0), va(0.0), vb(0.0), vac(0.0), vbc(0.0)
// コンストラクタ
{
  statv<double> statvi(r);
  dsmap.insert(make_pair(dose, statvi));
}

const gstat& gstat::operator=(const gstat& gstat0) // コピー代入演算子
{
  if (&gstat0 == this) return *this; // 自分自身への代入を防ぐ
  dsmap = gstat0.dsmap;  dsitrvec = gstat0.dsitrvec;  dstatc = gstat0.dstatc;
  vnd = gstat0.vnd;  vsa = gstat0.vsa;  vsb = gstat0.vsb;
  va = gstat0.va;  vb = gstat0.vb;
  vac = gstat0.vac;  vbc = gstat0.vbc;
  return *this;                 // 自分自身への参照を返す
}

const gstat& gstat::operator+=(const gstat& gstat0) // 加算演算子
{
  if (gstat0.nd() < 1) return *this; // 加算しない
  dstatc += gstat0.dstatc;  vnd += gstat0.vnd;  vsa += gstat0.vsa;
  return *this;                 // 自分自身への参照を返す
}

void gstat::sigma(double dose, double r) // 用量と反応データを累積する
{
  mapds_itr ditr = dsmap.find(dose);
  if (ditr == dsmap.end())      // 新規用量
  {
    statv<double> statvi(r);
    dsmap.insert(make_pair(dose, statvi));
  }
  else                      // すでにある用量
  { ditr->second.sigma(r);  }
  dstatc.sigma(dose, r);
}

int gstat::set()                // 用量別統計量を整理する
{
  vnd = 0;  vsa = 0.0;
  if (dsmap.empty()) return 0;  // 用量別統計量なし
  dsitrvec.resize(dsmap.size());
  for (mapds_itr ditr = dsmap.begin(); ditr != dsmap.end(); ++ditr, ++vnd)
  {
    dsitrvec[vnd] = ditr;
    vsa += ditr->second.n() * ditr->second.mean() * ditr->second.mean();
  }
  return vnd;                   // 用量数を返す
}

void gstat::cal()               // 群別回帰直線を計算する
{
  const double Veps = DBL_EPSILON * fmax(5.0, dstatc.syy());
  vsa -= dstatc.n() * dstatc.my() * dstatc.my();
  if (vsa < Veps) vsa = 0.0;
  double sxx = dstatc.sxx(), sxy = dstatc.sxy();
  vsb = (sxx > 0.0) ? sxy * sxy / sxx : 0.0;
  vb = (sxx > 0.0) ? sxy / sxx : 0.0;
  va = dstatc.my() - vb * dstatc.mx();
}

using mapsg = map<string, gstat>;         // 群別統計量マップの別名定義
using mapsg_itr = mapsg::iterator;        // 反復子の別名定義
using mapsg_citr = mapsg::const_iterator; // const反復子の別名定義

class varclass                  // 変数クラス
{
private:
  int vng;                      // 群数
  mapsg vgsmap;                 // 群別統計量マップ
  gstat gstatall;               // 全体の統計量用
  mat<double> smat;             // 回帰分析用行列
  double vsxxc;                 // 全体の共通用量平方和
  double vsp;                   // 全体の群差平方和
  double vsb2;                  // 全体の共通回帰平方和
  double vsid;                  // 全体の切片差平方和
  double vsbi;                  // 全体のブランク差平方和
  int vdfbi;                    // 全体のブランク差自由度
  double vsr;                   // 全体の残差平方和
  int vdfr;                     // 全体の残差自由度
  void cal_slope();             // 勾配比検定用計算
  void reg_mat();               // 行列を利用した回帰直線の計算
public:
  varclass() : vng(0), vgsmap(), gstatall(), smat(), vsxxc(0.0), vsp(0.0), vsb2(0.0), vsid(0.0), vsbi(0.0), vdfbi(0), vsr(0.0), vdfr(0) // デフォールトコンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int ng() const                // 群数を返す
  { return vng; }
  const mapsg& gsmap() const    // 群別統計量マップを返す
  { return vgsmap; }
  double ss(size_t iidx, size_t jidx) const // 積和行列の積和を返す
  { return (iidx < smat.r() && jidx < smat.c()) ? smat(iidx, jidx) : 0.0; }
  int n() const                 // 全体の例数を返す
  { return gstatall.n(); }
  int nd() const                // 全体の用量数を返す
  { return gstatall.nd(); }
  double mx() const             // 全体の用量平均を返す
  { return gstatall.mx(); }
  double my() const             // 全体の反応平均を返す
  { return gstatall.my(); }
  double sxx() const            // 全体の用量積和を返す
  { return gstatall.sxx(); }
  double syy() const            // 全体の総平方和を返す
  { return gstatall.syy(); }
  double sxy() const            // 全体の用量と反応の積和を返す
  { return gstatall.sxy(); }
  double a() const              // 全体の反応直線の定数を返す
  { return gstatall.a(); }
  double b() const              // 全体の反応直線の回帰係数を返す
  { return gstatall.b(); }
  double ac() const             // 全体の共通反応直線の定数を返す
  { return gstatall.ac(); }
  double bc() const             // 全体の共通反応直線の回帰位数を返す
  { return gstatall.bc(); }
  double sa() const             // 全体の用量平方和を返す
  { return gstatall.sa(); }
  double sb() const             // 全体の回帰平方和を返す
  { return gstatall.sb(); }
  double sxxc() const           // 全体の共通用量平方和を返す
  { return vsxxc; }
  double sp() const             // 全体の群差平方和を返す
  { return vsp; }
  double sb2() const            // 全体の共通回帰平方和を返す
  { return vsb2; }
  double sid() const            // 全体の切片差平方和を返す
  { return vsid; }
  double sbi() const            // 全体のブランク差平方和を返す
  { return vsbi; }
  int dfbi() const              // 全体のブランク差自由度を返す
  { return vdfbi; }
  double sr() const             // 全体の残差平方和を返す
  { return vsr; }
  int dfr() const               // 全体の残差自由度を返す
  { return vdfr; }
  int cal(const option& opt, ldata datalst);
  // データを読み込み勾配比検定を行う
};

void varclass::cal_slope()      // 勾配比検定用計算
{
  vsp = vsb2 = vsid =  vsbi = vsr = 0.0;  vdfbi = vdfr = 0;
  double s0r = 0.0, s1r = 0.0;
  int dfd = -1;
  mapsg_citr idx0 = vgsmap.end();
  for (mapsg_citr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr)
  {
    vsp += gitr->second.n() * gitr->second.my() * gitr->second.my();
    s0r += gitr->second.b() * gitr->second.sxy();
    if (gitr->second.nd() == 1 && gitr->second.dose(0) == 0.0) idx0 = gitr;
    dfd += gitr->second.nd();
  }
  vsp -= gstatall.n() * gstatall.my() * gstatall.my();
  const double Seps = DBL_EPSILON * fmax(5.0, gstatall.syy());
  if (vsp < Seps) vsp = 0.0;
  s0r = gstatall.syy() - vsp - s0r;
  if (s0r < Seps) s0r = 0.0;
  double ty = gstatall.n() * gstatall.my();
  double ty2 = gstatall.syy() + gstatall.n() * gstatall.my() * gstatall.my();
  int yidx = vng + 1;
  if (idx0 != vgsmap.end())
  {
    smat.assign(0, 0, (double)(gstatall.n() - idx0->second.n()));
    smat.assign(0, yidx, ty - idx0->second.n() * idx0->second.my());
    smat.assign(yidx, yidx, ty2 - idx0->second.n() * idx0->second.my() * idx0->second.my());
    reg_mat();
    s1r = smat(yidx, yidx);  vdfbi = 1;
  }
  smat = 0.0;
  smat.assign(0, 0, (double)(gstatall.n()));
  smat.assign(0, yidx, ty);  smat.assign(yidx, yidx, ty2);
  reg_mat();
  if (vdfbi > 0) { vsid = s1r - s0r;  vsbi = smat(yidx, yidx) - s1r; }
  else { vsid = smat(yidx, yidx) - s0r;  vsbi = 0.0; }
  if (vsid < Seps) vsid = 0.0;
  if (vsbi < Seps) vsbi = 0.0;
  vsb2 = gstatall.syy() - smat(yidx, yidx);
  if (vsb2 < Seps) vsb2 = 0.0;
  vsr = gstatall.syy() - gstatall.sa();
  if (vsr < Seps) vsr = 0.0;
  vdfr = gstatall.n() - 1 - dfd;
  int i = 1;
  for (mapsg_itr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr, ++i)
  { gitr->second.ac(smat(0, yidx));  gitr->second.bc(smat(i, yidx)); }
}

void varclass::reg_mat()        // 行列を利用した回帰直線の計算
{
  int yidx = vng + 1, i = 1;
  int n0 = (int)(smat(0, 0));
  for (mapsg_citr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr, ++i)
  {
    for (int j = 0; j < gitr->second.nd(); ++j)
    {
      int n = gitr->second.n(j);
      double x = gitr->second.dose(j);
      double y = gitr->second.n(j) * gitr->second.mean(j);
      smat.assign(0, i, smat(0, i) + n * x);
      smat.assign(i, i, smat(i, i) + n * x * x);
      smat.assign(i, yidx, smat(i, yidx) + x * y);
    }
  }
  for (int i = 1; i <= yidx; ++i)
  { smat.assign(i, 0, smat(0, i));  smat.assign(yidx, i, smat(i, yidx)); }
  smat.inv(vng + 1);
}

int varclass::cal(const option& opt, ldata datalst)
// データを読み込み勾配比検定を行う
{
  for (ldata_itr ditr = datalst.begin(); ditr != datalst.end(); ++ditr)
  {
    double dose = ditr->fdata(1), r = ditr->fdata(2);
    if (ditr->sdata(0) != BLANKs && dose != BLANKf && dose >= 0.0 && r != BLANKf)
    {
      mapsg_itr gitr = vgsmap.find(ditr->sdata(0));
      if (gitr == vgsmap.end()) // 新規群
      {
        gstat gstati(dose, r);
        vgsmap.insert(make_pair(ditr->sdata(0), gstati));
      }
      else                      // すでにある群
      { gitr->second.sigma(dose, r);  }
    }
  }
  for (mapsg_itr gitr = vgsmap.begin(); gitr != vgsmap.end(); ++gitr)
  {
    gitr->second.set();
    gstatall += gitr->second;
    gitr->second.cal();
  }
  gstatall.cal();
  vng = vgsmap.size();
  smat.resize(vng + 2, vng + 2);
  cal_slope();
  return vng;                   // 群数を返す
}

class slopeclass                // slopeメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_dstat(ostream& fout, const idd& id, const varclass& var) const;
  // 群ごとの基礎統計量と用量-反応直線を出力する
  void put_mstat(ostream& fout, const gstat& gstati) const;
  // 群ごとの用量別基礎統計量を出力する
  void put_erest(ostream& fout, const option& opt, const varclass& var) const;
  // 効力比推定値を出力する
  void put_slopetab(ostream& fout, const option& opt, const varclass& var) const;
  // 勾配比検定法用分散分析表を出力する
public:
  slopeclass() : fname()        // コンストラクタ
  {}
  ~slopeclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み保存・累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void slopeclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "GROUP ITEM   :" << id.item(0) << "\n";
  fout << "DOSE ITEM    :" << id.item(1) << "\n";
  fout << "RESPONSE ITEM:" << id.item(2) << "\n";
}

void slopeclass::put_dstat(ostream& fout, const idd& id, const varclass& var) const
// 群ごとの基礎統計量と用量-反応直線を出力する
{
  int i = 1;
  for (mapsg_citr gitr = (var.gsmap()).begin(); gitr != (var.gsmap()).end(); ++gitr, ++i)
  {
    fout << "\nGROUP" << setw(2) << i;
    fout << ":" << id.item(0) << "=" << gitr->first << "\n";
    put_mstat(fout, gitr->second);
    fout << "Group Dose-Response Line:y=" << gitr->second.a();
    fout << showpos << gitr->second.b() << noshowpos << "x\n";
    fout << "Common Dose-Response Line:y=" << gitr->second.ac();
    fout << showpos << gitr->second.bc() << noshowpos << "x\n";
    double r2 = (gitr->second.sa() > 0.0) ? gitr->second.sb() / gitr->second.sa() : 0.0;
    fout << "Contribution Rate if Group Dose-Response Line:r-squared=" << r2 << "\n";
  }
}

void slopeclass::put_mstat(ostream& fout, const gstat& gstati) const
// 群ごとの用量別基礎統計量を出力する
{
  fout << "        Dose         n        mean          SD          SE\n";
  constexpr int LENline = 58;
  putline(fout, LENline);
  for (int j = 0; j < gstati.nd(); ++j)
  {
    fout << setw(12) << gstati.dose(j) << setw(10) << gstati.n(j);
    fout << setw(12) << gstati.mean(j) << setw(12) << gstati.sd(j);
    fout << setw(12) << gstati.se(j) << "\n";
  }
  putline(fout, LENline);
  fout << "mean" << setw(8) << gstati.mx() << setw(10) << gstati.n();
  fout << setw(12) << gstati.my() << setw(12) << gstati.sdy();
  fout << setw(12) << gstati.sey() << endl;
}

void slopeclass::put_erest(ostream& fout, const option& opt, const varclass& var) const
// 効力比推定値を出力する
{
  fout << "\n<Ponency Ratio R and " << opt.cc() * 100.0 << "%CI";
  fout << "(Smaller Group Number is set as 1)>\n";
  double vr = (var.dfr() > 0) ? var.sr() / var.dfr() : 0.0;
  double t = ptot(1.0 - opt.cc(), var.dfr());
  mapsg_citr gitri = (var.gsmap()).begin();
  mapsg_citr gitrlast = (var.gsmap()).end();
  --gitrlast;
  int i = 1;
  while (gitri != gitrlast)
  {
    double b2s = gitri->second.bc(), css = var.ss(i, i);
    if (b2s != 0.0 && css != 0.0)
    {
      map<string, gstat>::const_iterator gitrj = gitri;
      ++gitrj;
      int j = i + 1;
      while (gitrj != (var.gsmap()).end())
      {
        double r = gitrj->second.bc() / b2s;
        double cuu = var.ss(j, j), csu = var.ss(i, j);
        fout << "R(GROUP" << j << "/GROUP" << i << ")=" << r;
        double g = t * t * vr * css / b2s / b2s;
        if (g >= 1.0) g = 0.0; // g>1だと計算できないため
        double gr = 1.0 - g;
        double v = vr * (cuu - 2.0 * r * csu + r * r * css - g * (cuu - csu * csu / css));
        double ci = t / b2s * sqrt(v);
        double rl = (r - g * csu / css - ci) / gr, ru = (r - g * csu / css + ci) / gr;
        fout << "    CI:Lower=" << rl << "  Upper=" << ru << "\n";
        ++gitrj;  ++j;
      }
    }
    ++gitri;  ++i;
  }
}

void slopeclass::put_slopetab(ostream& fout, const option& opt, const varclass& var) const
// 勾配比検定法用分散分析表を出力する
{
  if (var.syy() <= 0.0) return; // 出力なし
  constexpr int LENline = 78;
  int dfp = var.ng() - 1, dfd = -1;
  int dft = var.n() - 1, dfb = 1;
  for (mapsg_citr gitr = (var.gsmap()).begin(); gitr != (var.gsmap()).end(); ++gitr)
  { dfd += gitr->second.nd(); }
  string tstr("ANOVA Table");
  fout << "\n" << setw((LENline - strclen(tstr, CHARwide)) / 2 + tstr.size()) << tstr;
  fout << "\nFACTOR    Sum of Squares        DF  Mean Squares     F-value       p-value\n";
  putline(fout, LENline);
  double v = 0.0, f = 0.0, pr = 0.0;
  double vr = calvfp(v, f, pr, var.sp(), dfp, var.sr(), var.dfr());
  fout << "GROUP" << setw(19) << var.sp() << setw(10) << dfp << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  double sla = 0.0;
  int dfla = 0, i = 1;
  const double Seps = DBL_EPSILON * fmax(5.0, var.syy());
  mapsg_citr gitr = (var.gsmap()).begin();
  if (var.dfbi() > 0) { ++gitr;  ++i; }
  while (gitr != (var.gsmap()).end())
  {
    double sb = gitr->second.sb();
    calvfp(v, f, pr, sb, dfb, var.sr(), var.dfr());
    fout << "GROUP" << setw(2) << i << ":REG" << setw(13) << sb << setw(10) << dfb;
    fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
    int dfl = gitr->second.nd() - 1 - dfb;
    double sl = gitr->second.sa() - sb;
    if (dfl < 1 || sl < Seps) sl = 0.0;
    calvfp(v, f, pr, sl, dfl, var.sr(), var.dfr());
    fout << "GROUP" << setw(2) << i << ":LOF" << setw(13) << sl << setw(10) << dfl;
    fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
    dfla += dfl;  sla += sl;
    ++gitr;  ++i;
  }
  putline(fout, LENline);
  int dfid = dfp - var.dfbi();
  calvfp(v, f, pr, var.sid(), dfid, var.sr(), var.dfr());
  fout << "INTERCEPT" << setw(15) << var.sid() << setw(10) << dfid << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  int dfb2 = dfb * (var.ng() - var.dfbi());
  calvfp(v, f, pr, var.sb2(), dfb2, var.sr(), var.dfr());
  fout << "COMMON REG" << setw(14) << var.sb2() << setw(10) << dfb2 << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, sla, dfla, var.sr(), var.dfr());
  fout << "TOTAL LOF" << setw(15) << sla << setw(10) << dfla << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  calvfp(v, f, pr, var.sbi(), var.dfbi(), var.sr(), var.dfr());
  fout << "BLANK DIFF" << setw(14) << var.sbi() << setw(10) << var.dfbi() << setw(14) << v;
  fout << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  calvfp(v, f, pr, var.sa(), dfd, var.sr(), var.dfr());
  fout << "DOSE" << setw(20) << var.sa() << setw(10) << dfd;
  fout << setw(14) << v << setw(12) << f << setw(14) << pr << pind(pr) << "\n";
  fout << "RESUDIALS" << setw(15) << var.sr() << setw(10) << var.dfr() << setw(14) << vr << "\n";
  putline(fout, LENline);
  fout << "TOTAL" << setw(19) << var.syy() << setw(10) << dft << endl;
}

int slopeclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み保存・累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;
}

void slopeclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (opt.res("res") == false) return;
  put_id(cout, id);
  if (id.nc() < 1)              // データなし
  {
    cout << "\n" << endl;
    return;
  }
  varclass var;                 // 変数オブジェクト
  int ng = var.cal(opt, datalst);
  if (opt.res("stat")) put_dstat(cout, id, var);
  if (ng > 1 && opt.res("er")) put_erest(cout, opt, var);
  if (opt.res("at")) put_slopetab(cout, opt, var);
  cout << "\n" << endl;
}

int main(int argc, char *argv[]) // メインルーチン
{
  slopeclass slope;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) slope.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { slope.get(argv[i], opt, id, datalst);  fok = false; }
    }
    slope.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
